#!/bin/bash

help(){
    cat <<EOF
 Set up tool for KEGG FTP subscriber

 In this example, suppose KEGG data is downloaded under $HOME
 Set BIOROOT by entering

   on sh, bash, zsh:
        % BIOROOT=$HOME
        % export BIOROOT
        % PATH=$PATH:$BIOROOT/bin
   on csh, tcsh:
        % setenv BIOROOT $HOME
        % setenv PATH ${PATH}:$BIOROOT/bin

  % cd $HOME
  % /bin/ls -F
  README.kegg  RELEASE  brite/  dbget.6.5.tar.gz  genes/  ligand/  medicus/  module/  pathway/  xml/
  % tar xvzf dbget.6.5.tar.gz
  % cd dbget-6.5
  % ./install.sh

 By default, install.sh command will install DBGET package under BIOROOT.
EOF
    exit 0
}

export LANG=C

[ $# -gt 0 ] && help

dbget=$(pwd)
up=$(dirname $dbget)

read -p "Please input downloaded kegg directory ($up): " kegg
[ -z $kegg ] && kegg=$up
kegg=$(eval echo $kegg)
if [ -z $BIOROOT ]; then
    read -p "Please set BIOROOT ($kegg): " broot
    if [ -z $broot ]; then
        broot=$kegg
    fi
    broot=$(eval echo $broot)
    export BIOROOT=$broot
fi

date1=$(date +%m-%d)
date2=$(date +%y/%m)
date3=$(date +%y/%m/%d)
dt="-"
if [ -f $BIOROOT/RELEASE ]; then
    str=$(head -1 $BIOROOT/RELEASE)
    if [[ $str =~ ([0-9][0-9][0-9][0-9]\-[0-9][0-9]\-[0-9][0-9]) ]]; then
        dt=${BASH_REMATCH[1]}
        date1=$(date --date=$dt +%m-%d)
        date2=$(date --date=$dt +%y/%m)
        date3=$(date --date=$dt +%y/%m/%d)
    fi
else
    echo "Error: $BIOROOT/RELEASE was not found."
    exit 1
fi

read -p "Please input KEGG release number ($dt): " release
if [ -z $release ]; then
    release=$dt
fi

./configure --optimize=-O3
make

cd etc
echo "creating dbtab.."
if [ -f $kegg/genes/misc/taxonomy ]; then
    if [ ! -f $BIOROOT/etc/genestab -o \
           $BIOROOT/etc/genestab -ot $kegg/genes/misc/taxonomy ]; then
        cp $kegg/genes/misc/taxonomy .
        ./make_genestab
    fi
else
    echo "Error: taxonomy file was not found."
    exit 1
fi

cd $BIOROOT
mkdir -p bin etc man/man{1,5} db/kegg/{brite,disease,drug,genes,genome,ko,ligand,module,network,pathway}
cp $dbget/bin/{bget,bfind,btit,btab,seqnew,binfo} $BIOROOT/bin/
cp $dbget/etc/etc/{dbtab,keggtab,genestab,databases,databases-kegg,databases-genes,services,services-kegg,services-genes} $BIOROOT/etc/
cp $dbget/etc/etc/origfiles/genes.files $BIOROOT/etc/
cp $dbget/doc/*.1 $BIOROOT/man/man1/
cp $dbget/doc/*.5 $BIOROOT/man/man5/
$BIOROOT/bin/btab -p $BIOROOT/etc/dbtab -o $BIOROOT/etc/dbtab
seqnew=$BIOROOT/bin/seqnew
btab=$BIOROOT/bin/btab
dbinfotab=$BIOROOT/etc/dbinfotab.ideas
dbinfotab_up=0
makedb(){
    dbdir=$1
    db=$2
    dir=$3
    echo "$db..."
    cd $dbdir
    if [ -f $kegg/$dir/$db.gz ]; then
        gunzip -f $kegg/$dir/$db.gz
    fi
    if [ -f $kegg/$dir/$db ]; then
        ln -fs $kegg/$dir/$db .
    fi

    if [ -f $kegg/$dir/$db.ent.gz ]; then
        gunzip -f $kegg/$dir/$db.ent.gz
    fi
    if [ -f $kegg/$dir/$db.ent ]; then
        ln -fs $kegg/$dir/$db.ent $db
    fi

    if [ -f $kegg/$dir/$db.tar.gz ]; then
        run=1
        if [ -f $dbdir/$db.cdb ]; then
            if [ $kegg/$dir/$db.tar.gz -ot $dbdir/$db.cdb ]; then
                run=0
            fi
        fi
        if [ $run -eq 1 ]; then
            cd $kegg/$dir
            tar xzf $kegg/$dir/$db.tar.gz
            cd $dbdir
        fi
    fi
    if [ -f $kegg/$dir/$db/$db ]; then
        ln -fs $kegg/$dir/$db/$db .
        if [ -f $kegg/$dir/$db/${db}_ja ]; then
            ln -fs $kegg/$dir/$db/${db}_ja .
        fi
        if [ $db = "compound" ]; then
            ln -fs $kegg/$dir/$db/mol .
        elif [ $db = "drug" -o $db = "dgroup" ]; then
            ln -fs $kegg/$dir/$db/mol ${db}_mol
        fi
    fi

    if [ -f $dbdir/$db ]; then
        run=1
        if [ -f $dbdir/$db.cdb ]; then
            if [ $dbdir/$db -ot $dbdir/$db.cdb ]; then
                run=0
            fi
        fi
        if [ $run -eq 1 ]; then
            save_IFS=$IFS
            IFS=$'\n'
            entries=0
            residues=0
            for l in $($seqnew $db $db 2>&1); do
                if [[ $l =~ ^Number\ of\ entries\ +:\ ([0-9]+) ]]; then
                    entries=${BASH_REMATCH[1]}
                elif [[ $l =~ ^Number\ of\ residues\ +:\ ([0-9]+) ]]; then
                    residues=${BASH_REMATCH[1]}
                fi
                if [[ $l =~ STATISTICS|DATABASE|Number|Maximum ]]; then
                    echo $l
                fi
            done
            if [ $db = "ko" ]; then
                db="orthology"
            fi
            echo "$db:$release/$date1:$date2:$entries:$residues:$date3" >> $dbinfotab
            dbinfotab_up=1
            IFS=$save_IFS
        fi
    else
        echo "Error: $db was not found."
        exit 1
    fi
}

dbdir=$BIOROOT/db/kegg/genes
for sdir in genes/organisms genes/organisms_new; do
    for org in $(ls $kegg/$sdir/); do
        db=$($btab $org | cut -d " " -f 1)
        [ -z $db ] && continue;
        dir=$sdir/$org
        if [ ! -f $kegg/$sdir/$org/$db.ent.gz -a ! -f $kegg/$sdir/$org/$db.ent ]; then
            echo "Warning: $db.ent was not found in $dir."
            continue
        fi
        makedb $dbdir $db $dir
    done
done

dir=genes/addendum
db=T10000
makedb $dbdir $db $dir

dir=genes/viruses
db=T40000
makedb $dbdir $db $dir

genestit=$BIOROOT/db/kegg/genes/genes.tit
if [ ! -f $genestit -o $genestit -ot $dbinfotab ]; then
    echo "creating genes title.."
    cat /dev/null > $genestit
    for f in $(cat $BIOROOT/etc/genes.files); do
        cat $BIOROOT/db/kegg/genes/$f.tit >> $genestit
    done
    entries=$(wc -l $genestit | cut -d " " -f 1)
    echo "genes:$release/$date1:$date2:$entries::$date3" >> $dbinfotab    
fi

dir=genes
for db in ko genome; do
    dbdir=$BIOROOT/db/kegg/$db
    makedb $dbdir $db $dir
done

for db in brite module pathway; do
    dir=$db
    dbdir=$BIOROOT/db/kegg/$db
    makedb $dbdir $db $dir
done

dir=ligand
dbdir=$BIOROOT/db/kegg/$dir
for db in compound compound_ja enzyme glycan rclass reaction; do
    makedb $dbdir $db $dir
done

dir=medicus
dbdir=$BIOROOT/db/kegg/disease
for db in disease disease_ja; do
    makedb $dbdir $db $dir
done

dbdir=$BIOROOT/db/kegg/drug
for db in drug drug_ja dgroup dgroup_ja; do
    makedb $dbdir $db $dir
done

dbdir=$BIOROOT/db/kegg/network
for db in network variant; do
    makedb $dbdir $db $dir
done

if [ $dbinfotab_up -eq 1 ]; then
    echo "creating dbinfotab.."
    declare -A hash
    for l in $(cat $dbinfotab); do
        [[ $l =~ ([^:]+):(.+) ]]
        dname=${BASH_REMATCH[1]}
        dat=${BASH_REMATCH[2]}
        hash[$dname]=$dat
    done
    cat /dev/null > $dbinfotab
    for key in "${!hash[@]}"; do
        echo "$key:${hash[$key]}" >> $dbinfotab
    done
    sort $dbinfotab > $dbinfotab.tmp
    mv $dbinfotab.tmp $dbinfotab
fi

echo "Done."
echo "Please set following environment variables."
echo "export BIOROOT=$BIOROOT"
echo "export PATH=\$PATH:\$BIOROOT/bin"
