
#include "dbget.h"
#include "parseargs.h"
#include "readdbtab.h"
#include "netdbget.h"
#include "recdescent.h"
#include "makeskip.h"
#include "procdata.h"
#include "splitfile.h"
#include "bfind.h"

#define TOSTR2(n) #n
#define TOSTR(n) TOSTR2(n)

// fork() した順番で子プロセスの ID を保存
static list<pid_t> pidlist;
static map<pid_t,string> tempfiles;
static list<pid_t>::iterator pcitr;
static map<pid_t,string>::const_iterator fcitr;

static int style_option[2] = {0};

int main(int argc, char **argv)
{
  // 環境変数 BIOROOT が必須
  string bioroot;

#ifdef REST
  bioroot = "rest";
#endif

  char *bioenv = getenv("BIOROOT");
  if( bioenv && *bioenv )
    bioroot = bioenv;

  if( bioroot.empty() ){
    cerr << "BIOROOT : Environment variable undefined." << endl;
    return 1;
  }

  if( argc == 1 ){
    cerr << "bfind : Have to specify a database name." << endl;
    return 1;
  }

  // コマンドライン引数を取得する
  list<string> commandline; // netdbget() に渡すための文字列リスト
  list<string> arguments;
  for( int i = 1 ; i < argc ; i++ ){ // argv[0] is this program
    arguments.push_back( argv[i] );
  }
  commandline = arguments;

  {
    // commandline は netdbget に渡すため、単語中の文字としてスペースがあればエスケープする
    // エスケープしないとネットワーク越しに受け取った bfind が、単語の区切り文字として処理してしまう
    list<string>::iterator citr;
    string::size_type pos;
    string find_str = "\\"; // エスケープ文字をエスケープ
    string repl_str = "\\\\";
    for( citr = commandline.begin(); citr != commandline.end(); ++citr )
      for( pos = (*citr).find(find_str); pos != string::npos; pos = (*citr).find(find_str, repl_str.length() + pos) )
	(*citr).replace(pos, find_str.length(), repl_str);
    find_str = " "; // スペースをエスケープ
    repl_str = "\\ ";
    for( citr = commandline.begin(); citr != commandline.end(); ++citr )
      for( pos = (*citr).find(find_str); pos != string::npos; pos = (*citr).find(find_str, repl_str.length() + pos) )
	(*citr).replace(pos, find_str.length(), repl_str);
  }

  // 引数チェックと変数セット
  list<string> option_keys;
  option_keys.push_back("h");
  option_keys.push_back("l");
  option_keys.push_back("p");
  option_keys.push_back("t");
  option_keys.push_back("d");
  option_keys.push_back("P");

  multimap<string,string> options;
  list<string>::iterator aitr = arguments.begin();
  while( aitr != arguments.end() ){
    if( *aitr == "-a" ){
      style_option[0] = 1;
      aitr = arguments.erase(aitr);
      continue;
    }
    if( *aitr == "-n" ){
      style_option[1] = 1;
      aitr = arguments.erase(aitr);
      continue;
    }
    if( *aitr == "-C" ){
      options.insert(make_pair("C","case_sensitive"));
      aitr = arguments.erase(aitr);
      continue;
    }
    if( *aitr == "-W" ){
      options.insert(make_pair("W","word_search"));
      aitr = arguments.erase(aitr);
      continue;
    }
    ++aitr;
  }

  int r = parseargs(arguments,option_keys,options);
  if( r == 0 || r == -1 ) usage();
  if( r == 0 ) return 1;
  if( r < 0 ) return 0;

  // 引数がひとつの場合
  if( arguments.size() == 1 ){
    aitr = arguments.begin();
    string db = *aitr;
    string str;
    if( cin.bad() ){
      cerr << "Standard input was failed." << endl;
      return 1;
    }

    arguments.clear();

    list<string> lin;
    lci litr;
    while(1){

      // 標準入力が端末ならメッセージを表示
      if ( isatty( STDIN_FILENO ) )
	cerr << "Pattern  >> ";

      getline(cin,str);
      if( str.empty())
	return 0;

      splitToList(lin,str," ");
      litr = lin.begin();

      arguments.push_back(db);
      while( litr != lin.end() )
	arguments.push_back(*litr++);

      getdata(arguments, options, commandline, bioroot);
      arguments.clear();
    }

  } // if arguments == 1

  return getdata(arguments, options, commandline, bioroot);

}

int getdata(list<string>& arguments,
	    multimap<string,string>& options,
	    list<string>& commandline,
	    string& bioroot){

  bool verbose = false;
  bool debug   = false;
  bool case_sensitive = false;
  bool word_option = false;
  pair<sci,sci> optp;

  optp = options.equal_range("v");
  if( optp.first != optp.second )
    verbose = true;
  optp = options.equal_range("D");
  if( optp.first != optp.second )
    debug = true;
  optp = options.equal_range("C");
  if( optp.first != optp.second )
    case_sensitive = true;
  optp = options.equal_range("W");
  if( optp.first != optp.second )
    word_option = true;

  if( style_option[0] == 1 && style_option[1] == 1 ){
    cerr << "bfind : Cannot specify -a and -n at the same time." << endl;
    return 1;
  }

  // lオプションが指定されていれば取得する
  int displine = 0;  // lオプションで指定された表示行数
  int dispcount = 0; // 表示した行数
  optp = options.equal_range("l");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      displine = mystoi(pitr->second);
      pitr++;
    }
    if( displine < 1 ){
      usage();
      return 1;
    }
  }

  int MAXPROCS = 1;

#ifdef MAX_PROCS
  MAXPROCS = MAX_PROCS - 0;
#endif

  if( MAXPROCS < 1 ){
    usage();
    return 1;
  }

  // Pオプションが指定されていれば取得する
  optp = options.equal_range("P");
  if( optp.first != optp.second ){
    int proc_number = 1;
    sci pitr = optp.first;
    while(pitr != optp.second){
      proc_number = mystoi(pitr->second);
      pitr++;
    }
    if( proc_number < 1 ){
      usage();
      return 1;
    }
    if( proc_number > MAXPROCS ){
      usage();
      return 1;
    }
    if( proc_number < MAXPROCS )
      MAXPROCS = proc_number;
  }

  string dbtype;
  if( bioroot == "rest" ) dbtype = "url";
  optp = options.equal_range("t");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtype = pitr->second;
      pitr++;
    }
  }
  string dbpath;
  if( bioroot == "rest" ) dbpath = "http://rest.genome.jp/";
  optp = options.equal_range("d");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbpath = pitr->second;
      pitr++;
    }
  }
  if( dbtype.empty() && ! dbpath.empty() ){
    usage();
    return 1;
  }
  if( ! dbtype.empty() && dbpath.empty() ){
    usage();
    return 1;
  }

  list<string>::iterator aitr = arguments.begin();
  if( aitr == arguments.end() ){
    usage();
    return 1;
  }

  if( verbose || debug )
    while( aitr != arguments.end() )
      cout << *aitr++ << endl;

  // 引数の最初の単語はデータベース名
  aitr = arguments.begin();
  string db = *aitr;
  aitr = arguments.erase(aitr);

  // 残りの引数を式に変換
  list<string> keys;
  RecDescentParsing *rec = new RecDescentParsing();
  if( ! rec->parse(arguments,keys,debug) ){
    cerr << "Syntax error." << endl;
    return 1;
  }
  delete rec;

  // 検索語ごとにスキップテーブルを作成
  int skip[arguments.size()][256];
  int m = 0; // 評価式中の単語数
  int n = 0; // 検索語の数
  bool buffer_search = true; // ファイルを分割して並列検索
  makeskip(keys,m,n,buffer_search,case_sensitive,skip,debug);

  // l オプションが指定されていれば
  // ファイルを分割検索しない
  if( displine > 0 )
    buffer_search = false;

  // dbtab ファイルの設定
  // デフォルト dbtab == BIOROOT/etc/dbtab
  string dbtab;
  char *dbtabenv = getenv("DBTAB");
  if( dbtabenv && *dbtabenv ){
    dbtab = dbtabenv;
  }
  optp = options.equal_range("p");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtab = pitr->second;
      pitr++;
    }
  }
  if( dbtab.length() == 0 ){
    dbtab = bioroot;
    dbtab += DBTAB;
  }

  // file name check
  if( basename(dbtab).length() == 0 ){
    cerr << "Invalid Input File " << dbtab << endl;
    return 1;
  }

  // dbtab のデータリスト
  list<string> inputlist;

  string proxy;
  // t オプションと d オプションが指定されている場合
  if( ! dbtype.empty() && ! dbpath.empty() ){
    // 初期化
    inputlist.clear();
    if( bioroot == "rest" )
      inputlist.push_back(db + " " + dbpath + " 80");
    else
      inputlist.push_back(db + " " + dbtype + " " + dbpath + " " + dbtype);
  }
  // dbtab の情報を取得
  else if( ! readFile(dbtab,db,inputlist,proxy) ){
    cerr << db << ": No such database in dbtab." << endl;
    return 1;
  }

  // データベースファイルリスト
  list<string> filelist;

  // 省略名
  map<string,string> abbrmap;

  // 検索処理データリスト
  list<ProcData> plist;

  vector<string> lbuf;
  lci iitr = inputlist.begin();

  // fcount(fork count) と データベース名の対応
  map<int,string> forkdbmap;
  int cnt = 0;

  while( iitr != inputlist.end() ){

    // 半角スペース区切りの文字列をリストに変換
    splitToList(lbuf,*iitr++," ");

    vector<string>::const_iterator bitr = lbuf.begin();
    if( lbuf.size() == 4 ){
      // local database
      // ex.) /bio/db/ideas/genes/h.sapiens
      string fullpath = *(bitr+2) + "/" + *bitr;
      filelist.push_back(fullpath);
      abbrmap[fullpath] = *(bitr+3);

      // ファイルを分割する位置を設定
      splitfile(plist,fullpath,buffer_search);

    }

    else if( lbuf.size() == 3 ){
      // netdbget database
      // ex.) genbank dbget.genome.ad.jp 3266
      // ex.) prosite http://www.genome.jp/dbget-bin/dbgetserv? 80
      // netdbget(destinationhost,port,bfind,commandline,proxy);
      string srv = *(bitr+1);
      int port = mystoi(*(bitr+2));
      ProcData pdata(srv,port);
      plist.push_back(pdata);
    }

    // i-th fork process searches the database *bitr.
    int i = cnt;
    int psize = plist.size();
    for(i = cnt; i < psize; i++)
      forkdbmap[i] = *bitr;
    cnt = i;

  }

  lci kitr;
  lci kbeg = keys.begin();
  list<ProcData>::const_iterator itr;
  int procdata_size = plist.size();
  itr = plist.begin();
  if( procdata_size == 1 && buffer_search ){ // fork()+tempfileなし
    // 分割検索しないときに限り return 1 があり得る
    string abbr = abbrmap[(*itr).fullname];

    string server = (*itr).fullname;
    string::size_type rest = server.find("http://rest");
    if( rest != server.npos && rest == 0 )
      return netdbget((*itr).fullname,(*itr).port,"find",commandline,proxy) ? 0 : 1;
    else if( (*itr).port > 0 ) // netdbget database
      return netdbget((*itr).fullname,(*itr).port,"bfind",commandline,proxy) ? 0 : 1;
    else
      return search(kitr,kbeg,itr,abbr,displine,skip,m,buffer_search,word_option) ? 0 : 1;
  }

  // fork()+tempfileあり
  int fcount = 0; // fork() 呼び出し回数
  pid_t pid[procdata_size];
  string tempfile;

  // シグナルハンドラを設定
  const int signalList[] = {
    SIGHUP,
    SIGINT,
    SIGTERM,
    SIGQUIT,
    SIGPIPE,
  };
  const int signalListNum = sizeof signalList / sizeof signalList[0];
  struct sigaction sa;
  memset( &sa, 0, sizeof(struct sigaction) );  
  sa.sa_handler = clearTempFile;
  sa.sa_flags |= SA_RESTART;
  for( int i = 0; i < signalListNum; ++i ){
    if( sigaction( signalList[i], &sa, NULL ) != 0 ) {
      fprintf( stderr, "sigaction(2) error!\n" );
      exit(1);
    } 
  }

  for( itr = plist.begin(); itr != plist.end(); ++itr ){
    string abbr = abbrmap[(*itr).fullname];

    if( fcount >= MAXPROCS ){

      // fork() した順番で子プロセスの ID を取り出す
      pcitr = pidlist.begin();
      if( pcitr == pidlist.end() )
	break;

      // fork() した順番で子プロセスの終了処理
      if( waitpid(*pcitr,NULL,0) < 0 )
	break;

      // 子プロセスの出力した検索結果を格納した一時ファイル
      fcitr = tempfiles.find(*pcitr);
      if( fcitr == tempfiles.end() )
	break;

      // 検索結果を出力して一時ファイルを削除
      tempfile = fcitr->second;
      displayResult(tempfile,displine,dispcount);
      pidlist.erase(pcitr);

      // l オプションが指定されてた場合
      // 残りの子プロセスの終了処理
      if( displine > 0 && displine <= dispcount ){
	clearTempFile(0);
	break;
      }

    } // if fcount >= MAXPROCS

    // 出力バッファ用一時ファイル
    tempfile = setTempFileName();

    pid[fcount] = fork();
    if( pid[fcount] < 0 ) break;
    else if( pid[fcount] == 0 ){
      // 子プロセス用シグナルハンドラを設定
      sa.sa_handler = clearTempFileForChild;
      for( int i = 0; i < signalListNum; ++i ){
	if( sigaction( signalList[i], &sa, NULL ) != 0 ) {
	  fprintf( stderr, "sigaction(2) error!\n" );
	  exit(1);
	} 
      }
      pidlist.clear();
      tempfiles.clear();
      pidlist.push_back(fcount+1);
      tempfiles.insert(make_pair(fcount+1,tempfile));

      // 検索処理
      ofstream ofs(tempfile.c_str());
      if( !ofs ){
	cerr << "Cannot open file: " << tempfile << endl;
	continue;
      }

      // netdbget() の引数 cmd を作る
      // 最初の単語はデータベース名
      list<string> cmd;
      cmd.push_back(forkdbmap[fcount]);

      // commandline の 2つめ以降の単語が検索語
      lci fitr = commandline.begin();
      fitr++;
      while(fitr != commandline.end())
	cmd.push_back(*fitr++);

      string server = (*itr).fullname;
      string::size_type rest = server.find("http://rest");
      if( rest != server.npos && rest == 0 )
	netdbget((*itr).fullname,(*itr).port,"find",cmd,proxy,ofs);
      else if( (*itr).port > 0 ) // netdbget database
	netdbget((*itr).fullname,(*itr).port,"bfind",cmd,proxy,ofs);
      else
	search(kitr,kbeg,itr,abbr,displine,skip,m,buffer_search,word_option,ofs);

      ofs.close();
      exit(0);
    }
    else{
      pidlist.push_back(pid[fcount]);
      tempfiles.insert(make_pair(pid[fcount],tempfile));
      ++fcount;
    }

  } // for( itr = plist.begin(); itr != plist.end(); ++itr )

  // fork() した順番で子プロセスの ID を取り出す
  for( pcitr = pidlist.begin(); pcitr != pidlist.end(); ++pcitr ){

    // fork() した順番で子プロセスの終了処理
    if( waitpid(*pcitr,NULL,0) < 0 )
      break;

    // 子プロセスの出力した検索結果を格納した一時ファイル
    fcitr = tempfiles.find(*pcitr);
    if( fcitr == tempfiles.end() )
      break;

    // 検索結果を出力して一時ファイルを削除
    tempfile = fcitr->second;
    displayResult(tempfile,displine,dispcount);

    // l オプションが指定されてた場合
    // 残りの子プロセスの終了処理
    if( displine > 0 && displine <= dispcount )
      clearTempFile(0);

  }

  return 0;

}

bool search(lci& kitr,lci& kbeg,
	    list<ProcData>::const_iterator& itr,
	    string& abbr,int displine,
	    int skip[][256], int m, bool buffer_search,
	    bool word_option)
{
  return search(kitr,kbeg,itr,abbr,displine,skip,m,buffer_search,word_option,cout);
}

bool search(lci& kitr,lci& kbeg,
	    list<ProcData>::const_iterator& itr,  
	    string& abbr, int displine,
	    int skip[][256], int m, bool buffer_search,
	    bool word_option, ostream& ofs)
{

  char buf[MAXSIZE+1] = {0};

  int rsize = 0;
  int ssize = 0; // 最後の改行コードから残りのサイズ

  int buf_len = 0;
  int buf_offset = 0;

  int i = 0;
  int j = 0;
  int k = 0; // 評価式中の単語数のカウンタ
  int l = 0; // 検索語数のカウンタ

  int pos = 0;
  int spos = 0;
  int epos = 0;
  int lsize = 0;

  int64_t total = 0;
  int64_t count = 0;

  // 検索ファイルの分割位置を求める
  int64_t offset = (*itr).offset;
  int64_t size = (*itr).size;

  string titlefile = (*itr).fullname;
  titlefile += ".tit";

  int fd;

#ifdef SOL
  fd = open64(titlefile.c_str(),O_RDONLY);
#else
  fd = open(titlefile.c_str(),O_RDONLY);
#endif
  if( fd < 0 ){
    cerr << "Cannot open file: " << titlefile << endl;
    return false;
  }

#ifdef SOL
  if( lseek64(fd,offset,SEEK_SET) < 0 ){
#else
  if( lseek(fd,offset,SEEK_SET) < 0 ){
#endif
    cerr << "Cannot seek() file: " << titlefile << endl;
    return false;
  }

  if( buffer_search ){

    while( total < size ){

      // 前回の残りを buf へコピー
      memcpy(buf,&buf[rsize],ssize);

      if( size - total > MAXSIZE - ssize )
	count = read(fd, buf + ssize, MAXSIZE - ssize);
      else
	count = read(fd, buf + ssize, size - total);

      if( count < 0 )
	break;

      rsize = count;  // read() が読み込んだサイズ
      rsize += ssize; // buf に格納されているサイズ
      total += count; // 読み込み済みサイズ

      // buf の中にある最後の改行コードを線形探索
      // rsize は buf の中にある最後の改行コードまでのサイズ
      // buf の先頭から rsize までを探索する
      // ssize は残りのサイズでこの領域は次回の探索に持ち越す
      for ( ssize = 0; ; --rsize, ++ssize ) {
	if ( ! rsize )
	  break;
	else if ( buf[rsize - 1] == '\n' )
	  break;
      }

      // buf について探索
      buf_len = rsize;
      buf_offset = 0;
      while(1){

	if( buf_len < 1 ) break;

	k = 0;
	l = 0;
	kitr = kbeg;

	// buf の中に検索語が見つかった場合
	// ヒットした位置は pos
	pos = boyer_moore(buf+buf_offset, (*kitr).c_str(),
			  buf_len, (*kitr).length(),
			  skip[l], word_option);

	if( pos > -1 ){
	  pos += buf_offset;

	  // ヒットした単語を含む行を一行取り出す
	  // buf の先頭行のとき spos には -1 が入る
	  spos = 0;
	  epos = 0;
	  for( spos = pos; ; --spos )
	    if( spos < 0 || buf[spos] == '\n' )
	      break;
	  for( epos = pos; ; ++epos )
	    if( buf[epos] == '\n' )
	      break;

	  lsize = epos - spos;
	  char line[lsize+1];
	  memcpy(line,&buf[spos+1],lsize);
	  line[lsize] = '\0';

	  // ヒットした単語を含む行について残りの検索語を探索
	  // m は評価式中の単語数
	  for( ++k,++kitr; k < m; ++k,++kitr ){
	    if( *kitr == ".and." ) continue;
	    ++l;
	    pos = boyer_moore(line, (*kitr).c_str(),
			      lsize, (*kitr).length(),
			      skip[l], word_option);
	    if( pos < 0 ) break;
	  }

	  if( pos > -1 ){ // すべての検索語にマッチした

	    if( style_option[0] == 1 ) // a option
	      convert_a(line,titlefile);

	    if( style_option[1] == 1 ) // n option
	      convert_n(line);
	    else{
	      // 半角スペースで区切られた最初の単語に : が含まれていないときは
	      // 先頭に 省略名: を付加する
	      i = 0;
	      j = 0;
	      for( ; i < lsize; ++i)
		if( line[i] == ' ' ) break;
	      for( ; j < i; ++j)
		if(line[j] == ':' ) break;
	      if( i == j )
		ofs << abbr << ":";
	    }

	    ofs << line;

	  }

	  // 次の探索領域のオフセット
	  buf_offset = epos + 1;
	  buf_len = rsize - buf_offset;

	} // if pos > -1

	else
	  break;

      } // while(1)

    } // while( total < size )

  }
  else{

    list<bool> results;
    list<bool>::iterator ritr;
    bool r1,r2;

    while( total < size ){

      // 前回の残りを buf へコピー
      memcpy(buf,&buf[rsize],ssize);

      if( size - total > MAXSIZE - ssize )
	count = read(fd, buf + ssize, MAXSIZE - ssize);
      else
	count = read(fd, buf + ssize, size - total);

      if( count < 0 )
	break;

      rsize = count;  // read() が読み込んだサイズ
      rsize += ssize; // buf に格納されているサイズ
      total += count; // 読み込み済みサイズ

      // 一行毎にデータを取り出して探索
      spos = 0;
      for( epos = 0; epos < rsize; ++epos ){

	// 改行コードまでを一行として一行毎にデータを取り出して解析
	if( buf[epos] == '\n' ){
	  lsize = epos + 1 - spos;

	  char line[lsize+1];
	  memcpy(line,&buf[spos],lsize);
	  line[lsize] = '\0';

	  results.clear(); // 初期化

	  // m は評価式中の単語数
	  for( k = 0, l = 0, kitr = kbeg; k < m; ++k,++kitr ){

	    if( *kitr == ".and." ){
	      if( results.empty() ) break;
	      r1 = results.back();
	      results.pop_back();
	      if( results.empty() ) break;
	      r2 = results.back();
	      results.pop_back();
	      results.push_back( r1 && r2 );
	      continue;
	    }
	    else if( *kitr == ".or." ){
	      if( results.empty() ) break;
	      r1 = results.back();
	      results.pop_back();
	      if( results.empty() ) break;
	      r2 = results.back();
	      results.pop_back();
	      results.push_back( r1 || r2 );
	      continue;
	    }
	    else if( *kitr == ".not." ){
	      if( results.empty() ) break;
	      r1 = results.back();
	      results.pop_back();
	      results.push_back( ! r1 );
	      continue;
	    }

	    pos = boyer_moore(line, (*kitr).c_str(),
			      lsize, (*kitr).length(),
			      skip[l], word_option);
	    ++l;

	    if( pos > -1 )
	      results.push_back(true);
	    else
	      results.push_back(false);

	  }

	  if( k == m && results.back() ){ // すべての評価式を処理

	    if( style_option[0] == 1 ) // a option
	      convert_a(line,titlefile);

	    if( style_option[1] == 1 ) // n option
	      convert_n(line);
	    else{
	      // 半角スペースで区切られた最初の単語に : が含まれていないときは
	      // 先頭に 省略名: を付加する
	      i = 0;
	      j = 0;
	      for( ; i < lsize; ++i)
		if( line[i] == ' ' ) break;
	      for( ; j < i; ++j)
		if(line[j] == ':' ) break;
	      if( i == j )
		ofs << abbr << ":";
	    }

	    ofs << line;

	  }

	  spos = epos + 1; // 次行の先頭位置

	} // if buf[epos] == '\n'

      } // for epos < rsize

      ssize = rsize - spos; // 残りのサイズ
      rsize = spos;   // 今回処理したサイズ

    } // while( total < size )

  } // buffer_search or not

  close(fd);
  return true;

}

// 一意な出力ファイル名を作成する
string setTempFileName() {

  char tempname[128] = {0};

#ifdef TMPDIR
  strcpy(tempname,TOSTR(TMPDIR));
  strcat(tempname,"/bfind_process_file.XXXXXX");
#else
  strcpy(tempname,"/tmp/bfind_process_file.XXXXXX");
#endif

  int fd;
  fd = mkstemp(tempname);
  if( fd == -1 )
    cerr << "Unable to open " << tempname << endl;
  close(fd);
  unlink(tempname);

  string spid = to_string(getpid());
  char pbuf[32] = {0};
  strcpy(pbuf,"_");
  strcat(pbuf,spid.c_str());

  strcat(tempname,pbuf);
  return string(tempname);
}

void displayResult(string& filename,int& displine,int& dispcount) 
{  

  // if specified -l option
  if( displine > 0 ){

    if( dispcount < displine ){

      ifstream ifs(filename.c_str());
      if( !ifs ){
	remove(filename.c_str());
	cerr << "Cannot open file: " << filename << endl;
	return;
      }
      string buf;
      while( getline(ifs,buf) ){
	cout << buf << endl;
	if( ++dispcount >= displine )
	  break;
      }
      ifs.close();

    }
  }
  else{

    // display all of file
    cat(filename);

  }

  remove(filename.c_str());

}

int boyer_moore
(const char *buf, const char *key, int buf_len,int key_len, int *skip, bool word_option)
{

  int i,j;
  unsigned char p,q;
  int flag_s = 0, flag_e = 0;

  i = key_len - 1;
  while( i < buf_len ) {

    j = key_len - 1;

    //for(p=buf[i],q=key[j]; p == q; p=buf[--i],q=key[--j]){
    for(p=buf[i],q=key[j]; skip[p] == skip[q]; p=buf[--i],q=key[--j]){

      // 検索語がヒットした
      if( j == 0 ){

	// 検索語を単語として検索するオプション
	if( word_option ){

	  // i はヒットした検索語の先頭が現れる位置

	  // 検索語の先頭が buf の先頭か、もしくは
	  // 検索語のひとつ前の文字が非英数なら単語の語頭とみなす
	  if( i == 0 )
	    flag_s = 1;
	  else if( ! isalnum(buf[i-1]) )
	    flag_s = 1;

	  // 検索語の末尾が buf 末尾か、もしくは
	  // 検索語のひとつ後の文字が非英数なら単語の語尾とみなす
	  if( i + key_len + 1 == buf_len )
	    flag_e = 1;
	  else if( ! isalnum(buf[i+key_len]) )
	    flag_e = 1;

	  // 検索語は単語としてヒットした
	  if( flag_s && flag_e )
	    return i;

	}
	// 検索語は文字列としてヒットした
	else
	  return i;

      }
    }

    if( skip[p] > key_len - j )
      i += skip[p];
    else
      i += key_len - j;

  }

  return -1;


}

void clearTempFile(int sig){

  // 残りの子プロセスの終了と処理
  for( pcitr = pidlist.begin(); pcitr != pidlist.end(); ++pcitr ){
    kill(*pcitr,SIGTERM);
    waitpid(*pcitr,NULL,0);
    fcitr = tempfiles.find(*pcitr);
    if( fcitr == tempfiles.end() ) continue;
    remove((fcitr->second).c_str());
  }

  exit(0);
}

void clearTempFileForChild(int sig){

  // 子プロセスの終了処理
  for( pcitr = pidlist.begin(); pcitr != pidlist.end(); ++pcitr ){
    fcitr = tempfiles.find(*pcitr);
    if( fcitr == tempfiles.end() ) continue;
    remove((fcitr->second).c_str());
  }

  exit(0);
}

void convert_a(char *line,const string& titlefile){
  string title = line;

  // 2つめの単語が [xxxxx] という形式だったら
  // xxxxx が Accession と仮定
  // なければデフォルト形式で表示
  list<string> wordlist;
  splitToList(wordlist,title," ");
  lci witr = wordlist.begin();

  if( ++witr != wordlist.end() ){
    string acc = *witr;
    string::size_type ret1 = acc.find("[");
    string::size_type ret2 = acc.find("]");

    if( ret1 != acc.npos && ret1 == 0 &&
	ret2 != acc.npos && ret2 == (acc.length()-1) ){
      acc = acc.substr(ret1+1,ret2-1);

      // この acc が Accession かどうかを確かめる
      // acc がインデックスのキーになっていれば Accession
      // そうでなければデフォルト形式で表示
      bool flg = false;
      string ac = acc;
      string dummy;
      int fd;
      struct cdb cdb;
      if( open_index(fd,&cdb,titlefile) )
	flg = get_value(&cdb,ac,dummy);
      close_index(fd,&cdb);

      if( flg ){
	// 最初の単語
	string id = title;
	string content;
	string::size_type ret = title.find(" ");
	if( ret != title.npos )
	  id = title.substr(0,ret);

	// 最初の単語がコロンを含んでいる場合
	ret = id.find(":");
	if( ret != id.npos && ret + 1 < id.length() )
	  content = title.substr(ret+1);
	else
	  content = title;
	title = acc;
	title += " ";
	title += content;
	strcpy(line,title.c_str());
      }
    }
  }
}

void convert_n(char *line){
  string title = line;
  // 最初の単語
  string id = title;
  string::size_type ret = title.find(" ");
  if( ret != title.npos )
    id = title.substr(0,ret);
  // 最初の単語がコロンを含んでいる場合
  ret = id.find(":");
  if( ret != id.npos && ret + 1 < id.length() ){
    title = title.substr(ret+1);
    strcpy(line,title.c_str());
  }
  // コロンを含んでいなければ何もしない
}
