
#include "dbget.h"
#include "parseargs.h"
#include "readdbtab.h"
#include "netdbget.h"
#include "bget.h"

int main(int argc, char **argv)
{
  // 環境変数 BIOROOT が必須
  string bioroot;

#ifdef REST
  bioroot = "rest";
#endif

  char *bioenv = getenv("BIOROOT");
  if( bioenv && *bioenv )
    bioroot = bioenv;

  if( bioroot.empty() ){
    cerr << "BIOROOT : Environment variable undefined." << endl;
    return 1;
  }

  // コマンドライン引数を取得する
  list<string> arguments;
  for( int i = 1 ; i < argc ; i++ ){ // argv[0] is this program
    arguments.push_back( argv[i] );
  }
  list<string> commandline = arguments;

  // 引数チェックと変数セット
  list<string> option_keys;
  option_keys.push_back("h");
  option_keys.push_back("p");
  option_keys.push_back("t");
  option_keys.push_back("d");

  /******** f オプション
  bget -f hsa:1234 -> アミノ酸配列と塩基配列
  bget -f -n a hsa:1234 -> アミノ酸配列
  bget -f -n n hsa:1234 -> 塩基配列
  bget -f k dr:D00002 -> KCFデータ
  bget -f m dr:D00002 -> mol ファイルを cat
  *********/

  bool aaseq = false;
  bool ntseq = false;
  bool kcf = false;
  bool mol = false;

  list<string> seq_numbers;
  list<string>::iterator aitr = arguments.begin();
  while( aitr != arguments.end() ){
    if( *aitr == "-f" ){
      aitr = arguments.erase(aitr);

      if( aitr == arguments.end() ){
        aaseq = true;
        ntseq = true;
	break;
      }

      if( *aitr == "k" ){
        kcf = true;
        aitr = arguments.erase(aitr);
      }else if( *aitr == "m" ){
        mol = true;
        aitr = arguments.erase(aitr);
      }else{
        aaseq = true;
        ntseq = true;
	if( *aitr == "-n" )
	  continue;
	++aitr;
      }
      continue;
    }

    if( *aitr == "-n" ){
      if( ! get_seq_numbers(arguments,aitr,seq_numbers) )
	return 1;
      continue;
    }
    ++aitr;
  }

  // -n オプションが指定されている場合
  if( ! seq_numbers.empty() ){
    if( kcf || mol ){
      usage();
      return 1;
    }
    // -f オプションとセットで指定されなければならない
    if( ! aaseq || ! ntseq ){
      usage();
      return 1;
    }
    aaseq = false;
    ntseq = false;
  }

  multimap<string,string> options;
  int r = parseargs(arguments,option_keys,options);
  if( r == 0 || r == -1 ) usage();
  if( r == 0 ) return 1;
  if( r < 0 ) return 0;

  // 引数がひとつまたはない場合
  if( arguments.size() < 2 ){

    bool flag = true;

    if( arguments.size() == 1 ){
      aitr = arguments.begin();
      string key = *aitr;
      // 文字列にコロンを含んでいる場合
      string::size_type ret = key.find(":");
      if( ret != key.npos && ret + 1 < key.length() ){
	flag = false;
      }
    }

    if( flag ){
      string str;
      if( cin.bad() ){
	cerr << "Standard input was failed." << endl;
	return 1;
      }

      // 標準入力が端末ならメッセージを表示
      if ( isatty( STDIN_FILENO ) ){
	while(1){
	  cerr << "Entry > ";
	  getline(cin,str);
	  list<string> lin;
	  splitToList(lin,str," ");
	  lci litr = lin.begin();
	  str = *litr;
	  if( str.empty() )
	    return 0;
	  arguments.push_back(str);
	  getdata(arguments, options, commandline, bioroot,
		  aaseq, ntseq, kcf, mol, seq_numbers);
	  arguments.pop_back();
	}
      }
      else{
	// EOFが来るまで読み込み
	while( getline(cin,str) ){
	  list<string> lin;
	  splitToList(lin,str," ");
	  lci litr = lin.begin();
	  str = *litr;
	  if( str.empty() )
	    break;
	  arguments.push_back(str);
	}
      }
    }

  }

  return getdata(arguments, options, commandline, bioroot,
		 aaseq, ntseq, kcf, mol, seq_numbers);

}

int getdata(list<string>& arguments,
	    multimap<string,string>& options,
	    list<string>& commandline,
	    string& bioroot,
	    bool aaseq,
	    bool ntseq,
	    bool kcf,
	    bool mol,
	    list<string>& seq_numbers){

  bool verbose = false;
  bool debug   = false;
  pair<sci,sci> optp;

  bool colon = false;

  optp = options.equal_range("v");
  if( optp.first != optp.second )
    verbose = true;
  optp = options.equal_range("D");
  if( optp.first != optp.second )
    debug = true;

  string dbtype;
  if( bioroot == "rest" ) dbtype = "url";
  optp = options.equal_range("t");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtype = pitr->second;
      pitr++;
    }
  }
  string dbpath;
  if( bioroot == "rest" ) dbpath = "http://rest.genome.jp/";
  optp = options.equal_range("d");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbpath = pitr->second;
      pitr++;
    }
  }
  if( dbtype.empty() && ! dbpath.empty() ){
    usage();
    return 1;
  }
  if( ! dbtype.empty() && dbpath.empty() ){
    usage();
    return 1;
  }

  list<string>::iterator aitr = arguments.begin();
  if( aitr == arguments.end() ){
    //usage();
    return 1;
  }

  if( verbose || debug )
    while( aitr != arguments.end() )
      cout << *aitr++ << endl;

  // 引数を db1:entry1 db2:entry2 ... の形式でリスト化
  string db;
  list<string> keys;
  aitr = arguments.begin();
  while( aitr != arguments.end() ){
    string key = *aitr;

    // 文字列にコロンを含んでいる場合
    string::size_type ret = key.find(":");
    if( ret != key.npos && ret + 1 < key.length() ){
      colon = true;
      keys.push_back(*aitr++);
      continue;
    }

    // コロンを使った場合、残りの文字列でもコロンを使わなければいけない
    if( colon ){
      cerr << *aitr++ << " : Syntax Error." << endl;
      cerr << "Usage : dbname:identifier" << endl;
      continue;
    }

    // 第一引数は DB名
    if( db.empty() ){
      db = *aitr++;
      continue;
    }

    // 第二引数以降はエントリ名
    key = db;
    key += ":";
    key += *aitr++;
    keys.push_back(key);
  }

  // dbtab ファイルの設定
  // デフォルト dbtab == BIOROOT/etc/dbtab
  string dbtab;
  char *dbtabenv = getenv("DBTAB");
  if( dbtabenv && *dbtabenv ){
    dbtab = dbtabenv;
  }
  optp = options.equal_range("p");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtab = pitr->second;
      pitr++;
    }
  }
  if( dbtab.length() == 0 ){
    dbtab = bioroot;
    dbtab += DBTAB;
  }

  // file name check
  if( basename(dbtab).length() == 0 ){
    cerr << "Invalid Input File " << dbtab << endl;
    return 1;
  }

  // db1:entry1 db2:entry2 ... を順番に処理
  lci kitr = keys.begin();
  if( kitr == keys.end() ){
    aitr = arguments.begin();
    cerr << *aitr << " : Syntax Error." << endl;
    cerr << "Usage : dbname:identifier" << endl;
    return 1;
  }

  list<string> inputlist;  // dbtab のデータリスト
  db = "";
  string proxy;
  bool gflag = false;  // global found flag
  while( kitr != keys.end() ){
    string key = *kitr++;
    string::size_type ret = key.find(":");
    string adb = key.substr(0,ret);
    if( adb.empty() )
      adb = key;

    // t オプションと d オプションが指定されている場合
    if( ! dbtype.empty() && ! dbpath.empty() ){
      // 初期化
      inputlist.clear();
      if( bioroot == "rest" )
	inputlist.push_back(adb + " " + dbpath + " 80");
      else
	inputlist.push_back(adb + " " + dbtype + " " + dbpath + " " + dbtype);
    }
    // dbtab の情報を取得
    else if( db.empty() || db != adb ){
      db = adb;
      if( ! readFile(dbtab,db,inputlist,proxy) ){
	cerr << db << ": No such database in dbtab." << endl;
	continue;
      }
    }

    bool flg = false; // found flag
    lci iitr = inputlist.begin();
    vector<string> lbuf;
    while( iitr != inputlist.end() ){
      // 半角スペース区切りの文字列をリストに変換
      // ex.) h.sapiens genes /bio/db/kegg/genes hsa
      splitToList(lbuf,*iitr++," ");
      vector<string>::const_iterator bitr = lbuf.begin();
      if( lbuf.size() == 4 ){
	// local database
	// インデックスの key は「省略名:エントリ番号」の形式
	//string ikey = *(bitr+3);
	//ikey += ":";
	//ikey += key.substr(ret+1,key.length()-ret);
	// インデックスの key はエントリ番号
	string ikey = key.substr(ret+1,key.length()-ret);

	// -f -n 1,2 or -f -n a,n 
	// PDB 以外は aa か nt
	if( ! seq_numbers.empty() && *(bitr+1) != "pdb" ){
	  list<string>::iterator snitr = seq_numbers.begin();
	  while( snitr != seq_numbers.end() ){
	    if( *snitr == "a" ){
	      if( *(bitr+1) != "sequences" ) // sequences には -f -n a はない
		aaseq = true;
	    }
	    else if( *snitr == "n" ){
	      if( *(bitr+1) != "sequences" ) // sequences には -f -n n はない
		ntseq = true;
	    }
	    else{
	      int num = mystoi(*snitr);
	      string type = check_seq_number(bitr,ikey,num);
	      if( type == "a" )
		aaseq = true;
	      else if( type == "n" )
		ntseq = true;
	    }
	    ++snitr;
	  }
	  if( ! aaseq && ! ntseq )
	    continue;
	}

	int mode = 0;
	if( aaseq )
	  mode = 1;
	if( ntseq )
	  mode = 2;
	if( aaseq && ntseq )
	  mode = 3;
	if( kcf )
	  mode = 4;
	if( mol )
	  mode = 5;
	if( *(bitr+1) == "pdb" && mode == 3 )
	  mode = 6;
 
	// 既に見つかった場合は終了
	if( flg )
	  break;

	// -f -n 1,2 pdb ....
	if( *(bitr+1) == "pdb" && ! seq_numbers.empty() ){

	  list<string>::iterator snitr = seq_numbers.begin();
	  while( snitr != seq_numbers.end() ){

	    bool sflg = true;
	    int len = (*snitr).length();
	    for( int i = 0; i < len; ++i )
	      if( ! isdigit( (*snitr)[i] ) )
		sflg = false;

	    if( *snitr == "a" )
	      sflg = true;

	    if( sflg ){
	      int pdb_chain_number = mystoi(*snitr); // > 0
	      mode = 10 * pdb_chain_number + 6; // 16 or 26 or 36...

	      flg = display(bitr,ikey,mode);

	    }

	    ++snitr;

	  }

	}
	else{
	  flg = display(bitr,ikey,mode);
	}

      }
      else if( lbuf.size() == 3 ){
	// netdbget database
	// ex.) genbank dbget.genome.ad.jp 3266
	// ex.) prosite http://www.genome.jp/dbget-bin/dbgetserv? 80
	// netdbget(destinationhost,port,bfind,commandline,proxy);

	// 〜 Call netdbget procedure 〜

	// REST の引数 cmd を作る
	// 最初の単語はデータベース名
	list<string> cmd;
	cmd.push_back(key);
	string nseq;
        bool aa = false;
        bool nt = false;
	if( ! seq_numbers.empty() ){ // -f -n ...
	  list<string>::iterator snitr = seq_numbers.begin();
	  while( snitr != seq_numbers.end() ){
	    if( nseq.empty() )
	      nseq = "-f+-n+";
	    else
	      nseq += ",";
	    nseq += *snitr;

            if (*snitr == "a")
              aa = true;
            else if (*snitr == "n")
              nt = true;

	    ++snitr;
	  }
	}
	if(aaseq&&ntseq) cmd.push_back("-f");
	else if(aa&&nt) cmd.push_back("-f");
        else if(aa) cmd.push_back("/aaseq");
        else if(nt) cmd.push_back("/ntseq");
	else if(kcf) cmd.push_back("/kcf");
	else if(mol) cmd.push_back("/mol");
	else if(!nseq.empty()) cmd.push_back(nseq);

	string server = *(bitr+1);
	string::size_type rest = server.find("http://rest");
	int port = mystoi(*(bitr+2));
	if( rest != server.npos && rest == 0 )
	  flg = netdbget(*(bitr+1),port,"get",cmd,proxy);
	else
	  flg = netdbget(*(bitr+1),port,"bget",commandline,proxy);

      }

    } // while inputlist

    if( !flg )
      cerr << key << " : No such data." << endl;
    else
      gflag = true;

  } // while keys

  if( gflag )
    return 0;
  return 1;

}


string check_seq_number(vector<string>::const_iterator& bitr,string& key,int seq_number){

  string dbname = *bitr;      // データベース名
  string dbtype = *(bitr+1);  // データベースタイプ
  string dbdir  = *(bitr+2);  // ディレクトリパス
  string abbr   = *(bitr+3);  // 省略名

  string flatfile;
  string seq_type;

  // _seq_key
  // ex.)_aaseq_hsa:1234
  string _aaseq_key = "_aaseq_";
  _aaseq_key += key;

  // _ntseq_key
  // ex.)_ntseq_hsa:1234
  string _ntseq_key = "_ntseq_";
  _ntseq_key += key;

  int64_t _seq_offset = 0;
  int64_t _seq_size = 0;

  // /bio/db/ideas/.../dbname.cdb を開く
  string index_fullpath = dbdir;
  index_fullpath += "/";
  index_fullpath += dbname;
  int fd;
  struct cdb cdb;
  struct mcdb m;
  struct mcdb_mmap map;
  bool mcdb = false;

#ifdef MCDB
  if (!open_index_mcdb(fd,&m,&map,index_fullpath,mcdb))
    if(!open_index(fd,&cdb,index_fullpath))
      return "false";
#else
  if(!open_index(fd,&cdb,index_fullpath))
    if (!open_index_mcdb(fd,&m,&map,index_fullpath,mcdb))
      return "false";
#endif

  int counter = 0;
  if (mcdb) {
    if (get_index_mcdb(&m,&map,_aaseq_key,flatfile,_seq_offset,_seq_size)) {
      ++counter;
      if (counter == seq_number) {
        seq_type = "a";
      }
    }
    if (get_index_mcdb(&m,&map,_ntseq_key,flatfile,_seq_offset,_seq_size)) {
      ++counter;
      if (counter == seq_number) {
        seq_type = "n";
      }
    }
    close_index_mcdb(fd,&map);
  }
  else {
    if (get_index(&cdb,_aaseq_key,flatfile,_seq_offset,_seq_size)) {
      ++counter;
      if (counter == seq_number) {
        seq_type = "a";
      }
    }
    if (get_index(&cdb,_ntseq_key,flatfile,_seq_offset,_seq_size)) {
      ++counter;
      if (counter == seq_number) {
        seq_type = "n";
      }
    }
    close_index(fd,&cdb);
  }
  
  return seq_type;
}

bool display(vector<string>::const_iterator& bitr,string& key,int mode){

  string dbname = *bitr;      // データベース名
  string dbtype = *(bitr+1);  // データベースタイプ
  string dbdir  = *(bitr+2);  // ディレクトリパス
  string abbr   = *(bitr+3);  // 省略名

  // molファイルを表示
  if( mode == 5 )
    return display_mol(bitr,key);

  // knapsack の kcf ファイルを表示
  if( mode == 4 && dbname == "knapsack" )
    return display_knapsack_kcf(bitr,key);

  string flatfile;
  int64_t offset = 0;
  int64_t size = 0;
  bool flg = false;

  // 配列データ処理用変数
  //int dbtype_mode = 0;

  // PDB チェーンデータ用特殊処理
  //if( mode > 5 ){
  //dbtype_mode = mode; // 6 or 16 or 26 or 36...
  //mode = 1;
  //}

  // _aaseq_key
  // ex.)_aaseq_hsa:1234
  string _aaseq_key = "_aaseq_";
  _aaseq_key += key;
  int64_t _aaseq_offset = 0;
  int64_t _aaseq_size = 0;
  bool _aaseq_flg = false;

  // _ntseq_key
  // ex.)_ntseq_hsa:1234
  string _ntseq_key = "_ntseq_";
  _ntseq_key += key;
  int64_t _ntseq_offset = 0;
  int64_t _ntseq_size = 0;
  bool _ntseq_flg = false;

  // _kcf_key
  // ex.)_kcf_dr:D00100
  string _kcf_key = "_kcf_";
  _kcf_key += key;
  int64_t _kcf_offset = 0;
  int64_t _kcf_size = 0;

  // kcf オプション
  // compound_ja => compound, drug_ja => drug
  if( mode == 4 ){
    string::size_type rret = dbname.rfind("_ja");
    if( rret != dbname.npos && rret + 3 == dbname.length() ){
      dbname.erase(rret);
    }
  }

  // /bio/db/ideas/.../dbname.cdb を開く
  string index_fullpath = dbdir;
  index_fullpath += "/";
  index_fullpath += dbname;
  int fd;
  struct cdb cdb;
  struct mcdb m;
  struct mcdb_mmap map;
  bool mcdb = false;

#ifdef MCDB
  if (!open_index_mcdb(fd,&m,&map,index_fullpath,mcdb))
    if (!open_index(fd,&cdb,index_fullpath))
      return false;
#else
  if (!open_index(fd,&cdb,index_fullpath))
    if (!open_index_mcdb(fd,&m,&map,index_fullpath,mcdb))
      return false;
#endif

  switch(mode){
  case 0 :
    if (mcdb)
      flg = get_index_mcdb(&m,&map,key,flatfile,offset,size);
    else
      flg = get_index(&cdb,key,flatfile,offset,size);
    break;
  case 1 :
    if (mcdb)
      _aaseq_flg = get_index_mcdb(&m,&map,_aaseq_key,flatfile,_aaseq_offset,_aaseq_size);
    else
      _aaseq_flg = get_index(&cdb,_aaseq_key,flatfile,_aaseq_offset,_aaseq_size);
    flg = _aaseq_flg;
    break;
  case 2 :
    if (mcdb)
      _ntseq_flg = get_index_mcdb(&m,&map,_ntseq_key,flatfile,_ntseq_offset,_ntseq_size);
    else
      _ntseq_flg = get_index(&cdb,_ntseq_key,flatfile,_ntseq_offset,_ntseq_size);
    flg = _ntseq_flg;
    break;
  case 3 :
    if (mcdb) {
      _aaseq_flg = get_index_mcdb(&m,&map,_aaseq_key,flatfile,_aaseq_offset,_aaseq_size);
      _ntseq_flg = get_index_mcdb(&m,&map,_ntseq_key,flatfile,_ntseq_offset,_ntseq_size);
    }
    else {
      _aaseq_flg = get_index(&cdb,_aaseq_key,flatfile,_aaseq_offset,_aaseq_size);
      _ntseq_flg = get_index(&cdb,_ntseq_key,flatfile,_ntseq_offset,_ntseq_size);
    }
    flg = _aaseq_flg || _ntseq_flg;
    break;
  case 4 :
    if (mcdb) {
      flg = get_index_mcdb(&m,&map,_kcf_key,flatfile,_kcf_offset,_kcf_size);
      if (flg)
        flg = get_index_mcdb(&m,&map,key,flatfile,offset,size);
    }
    else {
      flg = get_index(&cdb,_kcf_key,flatfile,_kcf_offset,_kcf_size);
      if (flg)
        flg = get_index(&cdb,key,flatfile,offset,size);
    }
    break;
  default:
    // PDB チェーンデータの場合
    if (mcdb) 
      _aaseq_flg = get_index_mcdb(&m,&map,_aaseq_key,flatfile,_aaseq_offset,_aaseq_size);
    else
      _aaseq_flg = get_index(&cdb,_aaseq_key,flatfile,_aaseq_offset,_aaseq_size);
    flg = _aaseq_flg;
    break;
  }

  if (mcdb) {
    close_index_mcdb(fd,&map);
  }
  else{
    close_index(fd,&cdb);
  }

  if (!flg)
    return false;

  // 配列を FASTA形式で表示するためにタイトル情報を取得
  string _aaseq_title;
  string _ntseq_title;
  if( 0 < mode && mode != 4 ){
    // /bio/db/ideas/.../dbname.tit.cdb を開く
    string titlefile = index_fullpath;
    titlefile += ".tit";

#ifdef MCDB
    if (!open_index_mcdb(fd,&m,&map,titlefile,mcdb))
      if (!open_index(fd,&cdb,titlefile))
        return false;
#else
    if (!open_index(fd,&cdb,titlefile))
      if (!open_index_mcdb(fd,&m,&map,titlefile,mcdb))
        return false;
#endif

    // dbname.tit.cdb は
    // key に対応するデータがタイトル情報になるように作成されている
    if (_aaseq_flg) {
      if (mcdb)
        flg = get_value_mcdb(&m,&map,key,_aaseq_title);
      else
        flg = get_value(&cdb,key,_aaseq_title);

      if (flg)
	_aaseq_title = ">" + _aaseq_title;
      else
	_aaseq_title = ">" + key;

      // GENES系のデータベースは末尾に (A) を付ける
      if (dbtype == "genes")
	_aaseq_title += " (A)";
    }
    if (_ntseq_flg) {
      if (mcdb)
        flg = get_value_mcdb(&m,&map,key,_ntseq_title);
      else
        flg = get_value(&cdb,key,_ntseq_title);

      if (flg)
	_ntseq_title = ">" + _ntseq_title;
      else
	_ntseq_title = ">" + key ;

      // GENES系のデータベースは末尾に (N) を付ける
      if (dbtype == "genes")
	_ntseq_title += " (N)";
    }

    if (mcdb)
      close_index_mcdb(fd,&map);
    else
      close_index(fd,&cdb);
  }

  // フラットファイル名のフルパス
  string fullpath = dbdir;
  fullpath += "/";
  fullpath += flatfile;

#ifdef SOL
  fd = open64(fullpath.c_str(),O_RDONLY);
#else
  fd = open(fullpath.c_str(),O_RDONLY);
#endif
  if( fd < 0 ){
    cerr << "Cannot open file: " << fullpath << endl;
    return false;
  }

  switch(mode){
  case 0 :
    // エントリデータを表示
    display_data(fd,offset,size);
    break;
  case 1 :
  case 2 :
  case 3 :
    if( _aaseq_flg ){
      // タイトル行を表示
      //cout << _aaseq_title << endl;

      // AASEQデータを表示
      display_seqdata(fd,_aaseq_title,_aaseq_offset,_aaseq_size,bitr,mode);


    }
    if( _ntseq_flg ){
      // タイトル行を表示
      //cout << _ntseq_title << endl;

      // NTSEQデータを表示
      display_seqdata(fd,_ntseq_title,_ntseq_offset,_ntseq_size,bitr,mode);

    }
    break;
  default:
    display_seqdata(fd,_aaseq_title,_aaseq_offset,_aaseq_size,bitr,mode);
    break;
  case 4 :
    // ファイルの先頭から offset バイトまでポインタを移動
#ifdef SOL
    lseek64(fd,offset,SEEK_SET);
#else
    lseek(fd,offset,SEEK_SET);
#endif

    // 先頭行を表示

    // 読み込みバッファ
    char buf[130] = {0};

    // read() が読み込んだバイト数
    int count = read(fd,buf,128);
    if( count < 0 )
      break;
    buf[--count] = '\0';

    // 最初の改行コードを探す
    // ret は改行コードの添字(位置)
    int length = strlen(buf);
    int ret = 0;
    for(ret = 0; ret < length; ++ret )
      if( buf[ret] == '\n' )
	break;
    buf[ret] = '\n';
    buf[++ret] = '\0';

    // 先頭行を表示
    if( write(1,buf,ret) < 0 )
      break;

    // KCFデータを表示
    display_data(fd,_kcf_offset,_kcf_size);
    break;
  }

  close(fd);
  return true;

}

// ファイルの先頭から offset バイト移動して
// そこから size バイト分のデータをそのまま表示
void display_data(int fd,const int64_t offset,const int64_t size){

  // 読み込みバッファ
  char buf[MAXSIZE+1] = {0};

  // read() が読み込んだバイト数
  int count = 0;

  // 読み込み済みバイト数
  int64_t total = 0;

  // ファイルの先頭から offset バイトまでポインタを移動
#ifdef SOL
  lseek64(fd,offset,SEEK_SET);
#else
  lseek(fd,offset,SEEK_SET);
#endif

  while( total < size ){

    if( size - total > MAXSIZE )
      count = read(fd,buf,MAXSIZE);
    else
      count = read(fd,buf,size - total);

    if( count < 0 )
      break;

    buf[count] = '\0';
    if( write(1,buf,count) < 0 )
      break;
    total += count;

  }

}

// ファイルの先頭から offset バイト移動して
// そこから size バイト分のデータを配列データとして取り出し、
// 配列データの左にあるスペースを削除して表示する
void display_seqdata(int fd,
		     const string& title,
		     const int64_t offset,
		     const int64_t size,
		     vector<string>::const_iterator& bitr,
		     int mode){

  // PDB チェーンデータ
  string pdb_chain_name;
  int pdb_chain_number = 0;
  int pdb_chain_counter = 0;
  bool pdb_disp_flg = false;

  // PIR 配列データ
  bool pir_flg = false;

  // EMBL 配列データ
  bool embl_flg = false;

  string dbtype = *(bitr+1);
  int dbtype_mode = -1;

  if( dbtype == "genes" )
    dbtype_mode = 0;
  else if( dbtype == "swissprot" )
    dbtype_mode = 1;
  else if( dbtype == "trembl" )
    dbtype_mode = 1;
  else if( dbtype == "epd" )
    dbtype_mode = 1;
  else if( dbtype == "prf" )
    dbtype_mode = 1;
  else if( dbtype == "pdbstr" )
    dbtype_mode = 1;
  else if( dbtype == "ncbi-cdd" )
    dbtype_mode = 1;
  else if( dbtype == "refseq" )
    dbtype_mode = 2;
  else if( dbtype == "pir" ){
    dbtype_mode = 2;
    pir_flg = true;
  }
  else if( dbtype == "genbank" )
    dbtype_mode = 2;
  else if( dbtype == "embl" ){
    dbtype_mode = 2;
    embl_flg = true;
  }
  else if( dbtype == "prodom" )
    dbtype_mode = 3;
  else if( dbtype == "sequences" )
    dbtype_mode = 3;
  else if( dbtype == "pdb" )
    dbtype_mode = mode;

  if( dbtype_mode < 0 ){
    cerr << dbtype << " : No such sequence database." << endl;
    return;
  }

  // PDB の配列データは3文字コードなので1文字コードへ変換して表示する
  // 変換テーブルを設定する
  map<string,string> aminocode;
  string convline;
  if( mode > 5 ){
    pdb_chain_number = (mode - 6)/10; // 0 or 1 or 2 or 3...
    setAminoCode(aminocode);
  }

  // PDB 以外のデータベースはここでタイトルを表示
  else
    cout << title << endl;

  // ファイルの先頭から offset バイトまでポインタを移動
#ifdef SOL
  lseek64(fd,offset,SEEK_SET);
#else
  lseek(fd,offset,SEEK_SET);
#endif

  // 各行を読み込み、検索
  char readbuf[MAXSIZE+1] = {0};
  char keepbuf[MAXSIZE+1] = {0};
  char searchbuf[2*(MAXSIZE+1)] = {0};
  char buf[2*(MAXSIZE+1)] = {0};

  // read() が読み込んだバイト数
  int count = 0;

  // 読み込み済みバイト数
  int64_t total = 0;

  int length;
  int ret;
  int pret;

  // 配列データファイルをバッファに読み込む
  while( total < size ){

    if( size - total > MAXSIZE )
      count = read(fd,readbuf,MAXSIZE);
    else
      count = read(fd,readbuf,size - total);

    if( count < 0 )
      break;

    readbuf[count] = '\0';
    total += count;

    if( keepbuf[0] != '\0' ){
      strcpy(searchbuf,keepbuf);
      strcat(searchbuf,readbuf);
    }
    else{
      strcpy(searchbuf,readbuf);
    }
    keepbuf[0] = '\0';

    // 最後の改行コードを探す
    // ret は改行コードの添字(位置)
    length = strlen(searchbuf);
    ret = 0;
    for(int i = 0; i < length; ++i )
      if( searchbuf[i] == '\n' )
	ret = i;

    //  最後の改行コードから残りを keep
    if( ret != 0 && ret+1 < length ){
      strcpy(keepbuf,&searchbuf[ret+1]);
      searchbuf[ret+1] = '\0';
    }

    // 改行コードまでを一行として一行ずつ表示
    length = strlen(searchbuf);
    ret = 0;
    pret = 0;
    while( ret+1 < length ){
      // 次の一行を探索
      pret = ret;
      for( ; ret+1 < length; ++ret )
	if( searchbuf[ret] == '\n' )
	  break;

      strncpy(buf,&searchbuf[pret],ret-pret);
      buf[ret-pret] = '\0';
      ++ret;

      string linedata = buf;
      int len = strlen(buf);

      // GENES
      if( dbtype_mode == 0 )
	linedata = trimLeftSpaces(linedata);

      // SWISSPROT、PRF、RefSeq...
      else if( dbtype_mode < 3 ){

	// 文字列中の半角スペースを取る
	trimSpaces(linedata);

	// 文字列中のスラッシュを取る
	if( pir_flg )
	  trimChars(linedata,"/");

	// RefSeq、PIR、GenBank、EMBL
	if( dbtype_mode == 2 ){
	  if( embl_flg ){
	    // 文字列末尾の数字を取る
	    len = linedata.length();
	    for(int i = --len; i > -1; --i)
	      if(isdigit(linedata[i]))
		linedata.erase(i,1);
	      else
		break;
	  }
	  else{
	    // 文字列先頭の数字を取る
	    for(int i = 0; i < len; ++i)
	      if(isdigit(linedata[0]))
		linedata.erase(0,1);
	      else
		break;
	  }
	}

      }

      // PDB
      else if( dbtype_mode > 5 ){

	// チェーン名
	if(  pdb_chain_name[0] != linedata[11] ){
	  pdb_chain_name = linedata[11];
	  ++pdb_chain_counter;

	  // 未表示として残っているデータを出力
	  if( pdb_disp_flg && ! convline.empty() ){
	    convline += "\n";
	    if( write(1,convline.c_str(),convline.length()) < 0 )
	      return;
	    convline.erase();
	  }

	  // このチェーンを表示するかどうかの判定
	  if( pdb_chain_number == 0 ||
	      pdb_chain_number == pdb_chain_counter ){
	    // タイトル表示
	    cout << title << " (" << pdb_chain_name << ")" << endl;
	    pdb_disp_flg = true;
	  }
	  else
	    pdb_disp_flg = false;

	}

	if( pdb_disp_flg ){

	  // アミノ酸 3文字コードを 1文字コードに変換し、
	  // これを 1行あたり 60文字で表示する
	  linedata = linedata.substr(19);
	  convline += convAminoCode(linedata,aminocode);
	  if( convline.length() >= 60 ){
	    convline.insert(60,"\n");
	    if( write(1,convline.c_str(),61) < 0 )
	      return;
	    convline = convline.substr(61);
	  }
	}

      }

      // PDB 以外のデータベース
      if( dbtype_mode < 4 ){

	// PIR の配列データは1行に30文字なので 1行あたり 60文字で表示する
	// PDB 用の変数 convline を借用
	if( pir_flg ){
	  convline += linedata;
	  if( convline.length() >= 60 ){
	    convline.insert(60,"\n");
	    if( write(1,convline.c_str(),61) < 0 )
	      return;
	    convline = convline.substr(61);
	  }
	}
	// PRODOM、sequences は FASTA 形式の配列データ
	else if( dbtype_mode == 3 ){
	  convline += linedata;
	  while( convline.length() >= 60 ){
	    convline.insert(60,"\n");
	    if( write(1,convline.c_str(),61) < 0 )
	      return;
	    convline = convline.substr(61);
	  }
	}
	// PIR 以外のデータベース
	else{
	  linedata += "\n";
	  if( write(1,linedata.c_str(),linedata.length()) < 0 )
	    return;
	}
      }

    } // while( ret+1 <= length )

  }  // while( total < size )

  // PDB の配列データまたは PIR の配列データ
  // または PRODOMの配列データ用の特殊処理
  if( pdb_disp_flg || pir_flg || dbtype_mode == 3 ){
    // 未表示として残っているデータを出力
    if( ! convline.empty() ){
      convline += "\n";
      if( write(1,convline.c_str(),convline.length()) )
	return;
      convline.erase();
    }
  }

}

// mol ファイルを cat する
bool display_mol(vector<string>::const_iterator& bitr,string& key){

  string base = *bitr;      // データベース名
  string dir  = *(bitr+2);  // ディレクトリパス

  // compound_ja => compound, drug_ja => drug
  // dgroup_ja => dgroup
  string::size_type rret = base.rfind("_ja");
  if( rret != base.npos && rret + 3 == base.length() )
    base.erase(rret);

  // mol ファイルは compound、drug、dgroup のみ存在する
  if( base != "compound" &&
      base != "drug" &&
      base != "dgroup" )
    return false;

  // エントリ番号取り出し
  // ex.) cpd:C00001 -> C00001
  string::size_type ret = key.find(":");
  string num = key.substr(ret+1,key.length()-ret-1);
  if( num.empty() )
    return false;

  // mol ファイルの番号は大文字
  // ex.) C00001
  Upper(num);

  // mol
  // ex.) /bio/db/ideas/ligand/mol/C00001.mol
  //      /bio/db/ideas/ligand/drug_mol/D00001.mol
  string mol = dir;
  mol += "/";
  if( base == "compound" )
    mol += "mol/";
  else if ( base == "drug" )
    mol += "drug_mol/";
  else if ( base == "dgroup" )
    mol += "dgroup_mol/";
  mol += num;
  mol += ".mol";

#ifdef SOL
  int fd = open64(mol.c_str(),O_RDONLY);
#else
  int fd = open(mol.c_str(),O_RDONLY);
#endif

  if( fd < 0 )
    return false;
  close(fd);

  // mol ファイルを全部表示
  cat(mol);

  return true;
}

// knapsack の kcf ファイルを cat する
bool display_knapsack_kcf(vector<string>::const_iterator& bitr,string& key){

  string base = *bitr;      // データベース名
  string dir  = *(bitr+2);  // ディレクトリパス

  // knapsack 以外は処理しない
  if( base != "knapsack" )
    return false;

  // エントリ番号取り出し
  // ex.) knapsack:C00000001 -> C00000001
  string::size_type ret = key.find(":");
  string num = key.substr(ret+1,key.length()-ret-1);
  if( num.empty() )
    return false;

  // kcf ファイルの番号は大文字
  // ex.) C00000001
  Upper(num);

  // kcf
  // ex.) /bio/db/ideas/knapsack/kcf/C00000001.kcf
  string kcf = dir;
  kcf += "/kcf/";
  kcf += num;
  kcf += ".kcf";

#ifdef SOL
  int fd = open64(kcf.c_str(),O_RDONLY);
#else
  int fd = open(kcf.c_str(),O_RDONLY);
#endif

  if( fd < 0 )
    return false;
  close(fd);

  // kcf ファイルを全部表示
  cat(kcf);

  return true;
}

// key を探索し、対応する値からフラットファイル名と
// オフセット値とサイズを取得する
bool get_index(struct cdb *cdbp,string& key,string& flatfile,
               int64_t& offset,int64_t& size) {
  // value;
  // ex.) dr:d00100 /bio/db/ideas/ligand/drug 156698 2554
  string value;
  if (!get_value(cdbp,key,value))
    return false;
  get_contents(flatfile,value,offset,size);
  return true;
}

bool get_index_mcdb(struct mcdb *mp,struct mcdb_mmap *mapp,string& key,
                    string& flatfile,int64_t& offset,int64_t& size) {
  string value;
  if (!get_value_mcdb(mp,mapp,key,value))
    return false;
  get_contents(flatfile,value,offset,size);
  return true;
}

void get_contents(string& flatfile,string &value,int64_t& offset,int64_t& size) {
  vector<string> buf;
  splitToList(buf,value," ");
  vector<string>::const_iterator citr = buf.begin();
  flatfile = *(citr+1);     // フラットファイル
  offset = atoll((*(citr+2)).c_str()); // オフセット値
  size = atoll((*(citr+3)).c_str());   // データサイズ
}

void setAminoCode(map<string,string>& table){
  table["ALA"] = "A";
  table["ASX"] = "B";
  table["CYS"] = "C";
  table["ASP"] = "D";
  table["GLU"] = "E";
  table["PHE"] = "F";
  table["GLY"] = "G";
  table["HIS"] = "H";
  table["ILE"] = "I";
  // J
  table["LYS"] = "K";
  table["LEU"] = "L";
  table["MET"] = "M";
  table["ASN"] = "N";
  // O
  table["PRO"] = "P";
  table["GLN"] = "Q";
  table["ARG"] = "R";
  table["SER"] = "S";
  table["THR"] = "T";
  // U
  table["VAL"] = "V";
  table["TRP"] = "W";
  table["UNK"] = "X";
  table["TYR"] = "Y";
  table["GLX"] = "Z";
}

string convAminoCode(const string& str,map<string,string>& table){
  string buf;
  list<string> lstr;
  splitToList(lstr,str," ");
  list<string>::const_iterator itr = lstr.begin();
  for( ; itr != lstr.end(); ++itr )
    if( ! table[*itr].empty() )
      buf += table[*itr];
    else
      buf += "X";
  return buf;
}

bool get_seq_numbers(list<string>& arguments,
		     list<string>::iterator& aitr,
		     list<string>& seq_numbers){

  aitr = arguments.erase(aitr);
  if( aitr == arguments.end() ){
    usage();
    return false;
  }

  list<string> tmplist;

  // -f -n a or -f -n 1,2...
  // -f -n オプションの引数チェックと重複除去
  Lower(*aitr);
  splitToList(tmplist,*aitr,",");
  list<string>::iterator snitr = tmplist.begin();
  map<string,int> sqnum;
  while( snitr != tmplist.end() ){
    if( *snitr != "a" && *snitr != "n" ){
      bool flg = true;
      int len = (*snitr).length();
      for( int i = 0; i < len; ++i )
	if( ! isdigit( (*snitr)[i] ) )
	  flg = false;

      int anum = mystoi(*snitr);
      if( anum <= 0 )
	flg = false;

      if( ! flg ){
	cerr << *aitr << " : Invalid option for -n." << endl;
	return false;
      }
    }
    sqnum[*snitr] = 1;
    ++snitr;
  }

  //seq_numbers.clear();
  map<string,int>::iterator mapitr = sqnum.begin();
  while( mapitr != sqnum.end() ){
    seq_numbers.push_back(mapitr->first);
    ++mapitr;
  }

  aitr = arguments.erase(aitr);

  return true;

}
