
#include "dbget.h"
#include "parseargs.h"
#include "readdbtab.h"
#include "btab.h"

using namespace std;

int main(int argc, char **argv)
{
  // 環境変数 BIOROOT が必須
  string bioroot;
  char *bioenv = getenv("BIOROOT");
  if( bioenv && *bioenv ){
    bioroot = bioenv;
  }else{
    cerr << "BIOROOT : Environment variable undefined." << endl;
    return 1;
  }

  if( argc == 1 ){
    usage();
    return 1;
  }

  // コマンドライン引数を取得する
  list<string> commandline;
  list<string> arguments;
  for( int i = 1 ; i < argc ; i++ ){ // argv[0] is this program
    arguments.push_back( argv[i] );
  }
  commandline = arguments;

  // 引数チェックと変数セット
  list<string> option_keys;
  option_keys.push_back("l");
  option_keys.push_back("p");
  option_keys.push_back("o");
  option_keys.push_back("cdb");
  option_keys.push_back("mcdb");

  multimap<string,string> options;
  int r = parseargs(arguments,option_keys,options);
  if( r == 0 || r == -1 ) usage();
  if( r == 0 ) return 1;
  if( r < 0 ) return 0;

  bool verbose = false;
  bool debug   = false;
  bool mcdb = false;
  
#ifdef MCDB
    mcdb = true;
#endif

  pair<sci,sci> optp;

  // cdb と mcdb は同時に指定できない
  optp = options.equal_range("cdb");
  if (optp.first != optp.second) {
    mcdb = false;
    optp = options.equal_range("mcdb");
    if (optp.first != optp.second) {
      usage();
      return 1;
    }
  }
  optp = options.equal_range("mcdb");
  if (optp.first != optp.second)
    mcdb = true;
  optp = options.equal_range("v");
  if (optp.first != optp.second)
    verbose = true;
  optp = options.equal_range("D");
  if (optp.first != optp.second)
    debug = true;
  if (debug)
    verbose = true;

  string lkey;
  optp = options.equal_range("l");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      lkey = pitr->second;
      pitr++;
    }
  }

  // l オプションとデータベース名は同時に指定できない
  list<string>::iterator aitr = arguments.begin();
  if( aitr != arguments.end() && ! lkey.empty() ){
    usage();
    return 1;
  }
  // データベース名は複数指定できない
  if( ++aitr != arguments.end() ){
    usage();
    return 1;
  }

  aitr = arguments.begin();
  // 探索するデータベース名を指定する場合は
  // コマンドライン引数か l オプションで指定される
  string dbname;
  if( aitr != arguments.end() )
    dbname = *aitr;
  else
    dbname = lkey;

  // 出力ファイル名を指定する場合は
  // o オプションで指定される
  string dbtabout;
  optp = options.equal_range("o");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtabout = pitr->second;
      pitr++;
    }
  }

  // o オプションとデータベース名は同時に指定できない
  if( ! dbtabout.empty() && ! dbname.empty() ){
    usage();
    return 1;
  }

  // o オプションかデータベース名かどちらかを指定しないといけない
  if( dbtabout.empty() && dbname.empty() ){
    usage();
    return 1;
  }

  // dbtab ファイルの設定
  // デフォルト dbtab == BIOROOT/etc/dbtab
  string dbtab;
  char *dbtabenv = getenv("DBTAB");
  if( dbtabenv && *dbtabenv ){
    dbtab = dbtabenv;
  }
  optp = options.equal_range("p");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtab = pitr->second;
      pitr++;
    }
  }

  // デフォルト dbtab == BIOROOT/etc/dbtab
  if( dbtab.length() == 0 ){
    dbtab = bioroot;
    dbtab += DBTAB;
  }

  // file name check
  if( basename(dbtab).length() == 0 ){
    cerr << "Invalid Input File " << dbtab << endl;
    return 1;
  }

  // 探索データベース名がある場合
  // そのデータベースの情報を出力して終了
  if( ! dbname.empty() ){

    // dbtab のデータリスト
    list<string> inputlist;

    // dbtab の情報を取得
    string proxy;
    if( ! readFile(dbtab,dbname,inputlist,proxy) ){
      cerr << dbname << " : No such database in dbtab." << endl;
      return 1;
    }

    // データベースの情報
    lci itr = inputlist.begin();

    // 複合データベースかどうかを調べる
    bool complex = false;
    if( ++itr != inputlist.end() )
      complex = true;

    // データベースの情報を出力
    itr = inputlist.begin();
    while( itr != inputlist.end() ){
      if( lkey.empty() ){
	cout << *itr << endl;
      }
      else{

	// l オプションが指定されたデータベースは
	// 複合データベースの場合だけ情報を出力する
	if( ! complex ) break;

	string val = *itr;
	int len = val.length();
	int pos = 0;
	for( int i = 0; i < len; ++i ){
	  if( val[i] == ' ' || val[i] == '\t' ){
	    if( pos != i ){
	      // get first word
	      val = val.substr(pos,i-pos);
	      cout << val << endl;
	      break;
	    }
	    pos = i+1;
	  }
	}
      }
      ++itr;
    }
      
    return 0;

  }

  // dbtab を読み込む
  // dbset は dbtab の情報のうち、リンク情報以外の情報を持つ
  // linkinfo は各データベースのリンク情報を持つ
  // namelist は dbtab に記載されるデータベースの順番を保持する
  // ここでは dbset 以外は使わない
  multimap<string,BasicInfo> dbset;
  multimap<string,string> linkinfo;
  list<string> namelist;
  if (readDbtab(dbtab,dbset,linkinfo,namelist)) {
    if (verbose)
      cout << "Writing to cdb." << endl;
    if (writeCDB(dbtabout,dbset,namelist,mcdb))
      return 0;
  }
  return 1;

}

bool writeCDB(const string& outf, 
	      multimap<string,BasicInfo>& dbset,
	      list<string>& namelist, bool& mcdb)
{
  string basefile = outf;
  basefile += ".out";

  ofstream ofs(basefile.c_str());
  if( !ofs ){
    cerr << "Cannot open file: " << basefile << endl;
    return false;
  }

  mci mitr; // iterator of dbset
  sci sitr; // iterator of linkinfo
  lci litr = namelist.begin();

  while( litr != namelist.end() ){
    string key = *litr;

    // get BasicInfo from dbset
    pair<mci,mci> mp = dbset.equal_range(key);
    //int dist = distance(mp.first,mp.second);
    mitr = mp.first;
    BasicInfo info = mitr->second;
    string linedata = key;
    if (mcdb) linedata.clear();

    while(mitr != mp.second){
      info = mitr->second;
      if (mcdb) {
        if (!linedata.empty()) {
          linedata += "\t";
        }
      }
      else {
        linedata += "\t";
      }
      // local database
      if( info.getPath().length() > 0 ){
	string name = info.getName();
	string nick = info.getNick();
	if( isTNumber(name) || isRGNumber(name) )
	  Upper(name);
	if( isTNumber(nick) || isRGNumber(nick) )
	  Upper(nick);
	linedata += name;
	linedata += " ";
	linedata += info.getType();
	linedata += " ";
	linedata += info.getPath();
	linedata += " ";
	linedata += nick;
      }
      // netdbget database
      else if( info.getHost().length() > 0 ){
	string name = info.getName();
	if( isTNumber(name) ||isRGNumber(name) )
	  Upper(name);
	linedata += info.getName();
	//linedata += " ";
	//linedata += "netdbget";
	linedata += " ";
	linedata += info.getHost();
	linedata += " ";
	linedata += itos(info.getPort());
      }
      mitr++;
    }

    if (mcdb) {
      ofs << "+" << key.length() << ","
          << linedata.length() << ":"
          << key << "->" << linedata
          << endl;
    }
    else {
      ofs << linedata << endl;
    }

    litr++;
  } // loop of namelist

  if (mcdb) {
    ofs << "+0,0:->" << endl;
    ofs << endl;
  }
  ofs.close();

  string cdbfile = outf;
  if (mcdb) {
    cdbfile += ".mcdb";
    if (!create_mcdb(basefile,cdbfile))
      return false;
  }
  else {
    cdbfile += ".cdb";
    if (!create_cdb(basefile,cdbfile))
      return false;
  }

  if (unlink(basefile.c_str()) < 0) {
    cerr << basefile << " : unlink error." << endl;
    return false;
  }

  return true;

}

bool create_cdb(const string& basefile,
		const string& cdbfile){

  string buf;
  list<string> lbuf;
  ifstream ifs(basefile.c_str());
  if( !ifs ){
    cerr << basefile << " : open error." << endl;
    return false;
  }

  int fd;
  struct cdb_make cdbm;
  unsigned keylen, datalen;
    
  // CDB を初期化
  fd = open(cdbfile.c_str(), O_RDWR|O_CREAT|O_TRUNC, 0644);
  cdb_make_start(&cdbm, fd);

  while( getline(ifs,buf) ) {

    // タブで分割した最初の単語をキーにする
    // データはキーを除いた残りの文字列
    splitToList(lbuf,buf,"\t");
    lci itr = lbuf.begin();
    keylen = (*itr).length();
    buf = buf.substr(keylen+1,buf.length() - keylen);

    datalen = buf.length();

    if (cdb_make_exists(&cdbm, (*itr).c_str(), keylen) == 0)
      cdb_make_add(&cdbm, (*itr).c_str(), keylen, buf.c_str(), datalen);
    // or use cdb_make_put() with appropriate flags
  }

  // finalize and close the database
  cdb_make_finish(&cdbm);
  close(fd);

  return true;
}

bool create_mcdb(const string& basefile,
                 const string& cdbfile){
  const char *fname = cdbfile.c_str();
  const char *input = basefile.c_str();
  if (mcdb_makefmt_fileintofile(input, fname, malloc, free) < 0) {
    cerr << "mcdb_makefmt_fileintofile() : Error." << endl;
    return false;
  }
  return true;
}

/*****
void writeFileB(const string& outf, 
		multimap<string,BasicInfo>& dbset,
		multimap<string,string>& linkinfo,
		list<string>& namelist)
{

  string output = outf + ".bin";
  if( verbose ){
    cout << "Writing to " << output << endl;
  }

  ofstream ofs(output.c_str());
  if( !ofs ){
    cerr << "Cannot open file: " << output << endl;
    exit(-1);
  }

  // dbtab.bfind and dbtab.bfind.bin
  string bfindtab = outf + ".bfind";
  string bfindbin = outf + ".bfind.bin";
  long offset; // offset of dbtab.bfind.bin

  ofstream oft(bfindtab.c_str());
  if( !oft ){
    cerr << "Cannot open file: " << bfindtab << endl;
    exit(-1);
  }
  ofstream ofb(bfindbin.c_str());
  if( !ofb ){
    cerr << "Cannot open file: " << bfindbin << endl;
    exit(-1);
  }

  // target files
  string gbfile = ".seq";
  string embfile = ".dat";
  string rsfile = ".gbff";

  mci mitr; // iterator of dbset
  sci sitr; // iterator of linkinfo
  lci litr = namelist.begin();
  while( litr != namelist.end() ){
    string key = *litr;

    // get BasicInfo from dbset
    pair<mci,mci> mp = dbset.equal_range(key);
    int dist = distance(mp.first,mp.second);
    mitr = mp.first;
    BasicInfo info = mitr->second;

    // get link from linkinfo
    pair<sci,sci> sp = linkinfo.equal_range(key);
    int lnum = distance(sp.first,sp.second);
    sitr = sp.first;
    string link = sitr->second;

    // gb or emb or rs
    int bflag = 0;
    if( key.find(gbfile)  == key.npos &&
	key.find(embfile) == key.npos &&
	key.find(rsfile)  == key.npos ){
      bflag = 1;
      offset = ofb.tellp();
    }

    ofs << key << " " << dist << " " << lnum << endl;

    while(mitr != mp.second){
      info = mitr->second;
      // local database
      if( info.getPath().length() > 0 ){
	ofs << info.getName() << " "
	    << info.getType() << " "
	    << info.getPath() << " "
	    << info.getNick() << endl;
	if( bflag ) // dbtab.bfind.bfin
	  ofb << info.getName() << " "
	      << info.getType() << " "
	      << info.getPath() << " "
	      << info.getNick() << endl;
      }
      // netdbget database
      else if( info.getHost().length() > 0 ){
	ofs << info.getName() << " "
	    << info.getHost() << " "
	    << info.getPort() << endl;
	if( bflag ) // dbtab.bfind.bfin
	  ofb << info.getName() << " "
	      << info.getHost() << " "
	      << info.getPort() << endl;
      }
      mitr++;
    }

    // link info
    while(sitr != sp.second){
      link = sitr->second;
      ofs << link << endl;
      sitr++;
    }

    // dbtab.bfind
    if( bflag ){
      oft << key << "\t"
	  << offset << "\t"
	  << dist << endl;
    }

    litr++;
  } // loop of namelist

  ofb.close();
  oft.close();
  ofs.close();
}
*****/
