
#include "dbget.h"
#include "refseq.h"

using namespace std;

Refseq::Refseq():
    _entry("LOCUS"),
    _ac("ACCESSION"),
    _de("DEFINITION"),
    _delim("//"),
    _title_ac(""),
    _title_de(""),
    // 配列データは塩基配列かアミノ酸配列の
    // いずれかのデータしか持たないため、
    // プログラム内部の変数名としてはいずれの場合でも
    // _aaseq で取り扱う
    _aaseq("ORIGIN"),
    _aaseq_id(""),
    _aaseq_offset(0),
    _aaseq_size(0),
    _aaseqlen(0),
    _flg_pep(false){}

Refseq::~Refseq(){}

void Refseq::parser(ifstream& ifs,
		    string& buf){

  if( buf.find(_entry) == 0 ){
    _header = _entry;
    set_offset(ifs,buf);

    // _id のセットは特殊処理 - アミノ酸配列のデータか塩基配列のデータかを判定しておく
    // set_id() は使わない
    // ex.)
    // LOCUS       NP_775166                543 aa            linear   ROD 11-FEB-2008
    vector<string> lbuf;
    splitToList(lbuf,buf," ");
    vector<string>::const_iterator itr = lbuf.begin();

    // ENTRY が2回以上は不正
    if( ! _id.empty() )
      ++_rejected;
    else{
      // 半角スペースで分割したときの 2つ目の単語を取り出す
      _id = *(itr+1);

      // tit ファイル用の ID
      _tit_id = _abbr;
      _tit_id += ":";
      _tit_id += _id;

      ++_counter;
    }

    // ここが aa だったらアミノ酸配列のデータ
    // それ以外は塩基配列のデータ
    if( *(itr+3) == "aa" )
      _flg_pep = true;

  }

  else if( buf.find(_de) == 0 ){
    _header = _de;
    buf = buf.substr(_header.length());
    _title_de = trimLeftSpaces(buf);
  }

  else if( buf.find(_ac) == 0 ){
    _header = _ac;
    buf = buf.substr(_header.length());
    buf = trimLeftSpaces(buf);

    // スペースで連結
    if( ! _accession.empty() )
      _accession += " ";
    _accession += buf;

    // タイトルに使用する accession はプライマリのみ
    if( _title_ac.empty() ){
      // 半角スペースで区切られた最初の単語
      string::size_type ret = buf.find(" ");
      if( ret != buf.npos )
	buf = buf.substr(0,ret);
      _title_ac = buf;
    }

  }

  else if( buf.find(_aaseq) == 0 ){
    _header = _aaseq;
    _aaseq_offset = ifs.tellg();
  }

  // 何かの継続行
  else if( buf.find(" ") == 0 ){
    if( _header == _de ){
      extend_title(_title_de,buf);
    }
    else if( _header == _ac ){

      // タイトルに使用するプライマリ accession は取得済み
      // スペースで連結
      _accession += " ";
      _accession += trimLeftSpaces(buf);
    }

    // 配列データの処理
    // ※ この統計処理は seqnew の実行結果に出るのみ
    else if( _header == _aaseq ){

      // ex.) 配列データのオフセットは 10バイト
      //        1 msyrwldiqa sqlrrmessi nhisqtvdih kekvarreig ilttnkntsr thkiiapasl
      //       61 erpvryirkp idytilddig hgvkvstqnm kmgglprttp ptqkppsppm sgkgtlgrhs
      buf = buf.substr(10);
      trimSpaces(buf);
      int len = buf.length();

      // このエントリの配列の長さ
      _aaseqlen += len;

      // データベース全体の配列の長さ
      _total_aaseqlen += len;

      // 最長の配列
      if( _max_aaseqlen < _aaseqlen )
	_max_aaseqlen = _aaseqlen;

      // 最長の配列(GENESならアミノ酸と塩基の配列の長さの合計)
      _max_seqlen = _max_aaseqlen;

    }
  }

  // エントリ終了処理
  else if( buf.find(_delim) == 0 ){
    set_size(ifs);

    // ここまでが ORIGIN行なら、ここで配列データの統計情報をセット
    if( _header == _aaseq )
      set_seq_size(buf,_aaseq_size,_aaseq_offset);

    writer();
  }

  // 何かのヘッダ行
  else{

    // ここまでが ORIGIN行なら、ここで配列データの統計情報をセット
    if( _header == _aaseq ){
      _size = ifs.tellg();
      _size -= _offset;
      set_seq_size(buf,_aaseq_size,_aaseq_offset);
    }

    set_header(buf);
  }

}

void Refseq::writer(){

  // プライマリ accession は大括弧で括る
  if( ! _title_ac.empty() ){
    _title = "[";
    _title += _title_ac;
    _title += "]";
  }

  // accession の後ろは半角スペースで連結
  if( ! _title.empty() && ! _title_de.empty() )
    _title += " ";

  _title += _title_de;

  if( _flg_pep )
    _aaseq_id = "_aaseq_";
  else
    _aaseq_id = "_ntseq_";
  _aaseq_id += _id;

  Seqnew::writer();

  if( _aaseq_offset > 0 )
    Seqnew::writer(_aaseq_id,_aaseq_offset,_aaseq_size);

  clear();
}

void Refseq::clear(){

  Seqnew::clear();

  _title_ac = "";
  _title_de = "";
  _aaseq_id = "";
  _aaseq_offset = 0;
  _aaseq_size = 0;
  _aaseqlen = 0;
  _flg_pep = false;

}
