
#include "dbget.h"
#include "seqnew.h"
#include "parseargs.h"
#include "readdbtab.h"

#include "genes.h"
#include "orthology.h"
#include "brite.h"
#include "pathway.h"
#include "module.h"
#include "network.h"
#include "variant.h"
#include "disease.h"
#include "enzyme.h"
#include "compound.h"
#include "dgroup.h"
#include "drug.h"
#include "glycan.h"
#include "reaction.h"
#include "rclass.h"
#include "aaindex.h"
#include "expression.h"
#include "pdbstr.h"
#include "ncbi_cdd.h"
#include "swissprot.h"
#include "refseq.h"
#include "pdb.h"
#include "carbbank.h"
#include "epd.h"
#include "pfam.h"
#include "pmd.h"
#include "prosite.h"
#include "prosdoc.h"
#include "sequences.h"
#include "genbank.h"
#include "embl.h"
#include "compound_ja.h"
#include "dgroup_ja.h"
#include "drug_ja.h"
#include "disease_ja.h"

int main(int argc, char **argv)
{

  // 環境変数 BIOROOT が必須
  string bioroot;
  char *bioenv = getenv("BIOROOT");
  if( bioenv && *bioenv ){
    bioroot = bioenv;
  }else{
    cerr << "BIOROOT : Environment variable undefined." << endl;
    return 1;
  }

  if( argc == 1 ){
    usage();
    return 1;
  }

  // コマンドライン引数を取得する
  list<string> commandline;
  list<string> arguments;
  for( int i = 1 ; i < argc ; i++ ){ // argv[0] is this program
    arguments.push_back( argv[i] );
  }
  commandline = arguments;

  // 引数がひとつもない場合
  if( arguments.size () == 0 ){
    usage ();
    return 1;
  }

  list<string>::iterator aitr = arguments.begin();

  char dinput[DBGET_BUFSIZ+1] = {0};
  bool seqnew_p_flg = false;

  // seqnew_p から呼び出された場合
  // 引数は数字ひとつでファイル記述子が渡される
  // このファイル記述子は親プロセス seqnew_p が開いたパイプ
  if( arguments.size () == 1 ){

    bool iflg = true;
    int len = (*aitr).length();
    for( int i = 0; i < len; ++i )
      if( ! isdigit( (*aitr)[i] ) )
	iflg = false;

    // 引数は数字
    if( iflg ){
      int fd = mystoi( *aitr );
      arguments.clear();
      int fcount = read(fd, dinput, DBGET_BUFSIZ);
      if( fcount == -1 )
	return 1;
    }

    // 環境変数 SEQNEWOPT が渡されている
    char *option = getenv("SEQNEWOPT");
    if( option != NULL ){

      seqnew_p_flg = true;

      int ioption = atoi(option);
      if( ! (ioption / 100 && ( ioption % 100 ) / 10) ){
	if( ioption / 100 )
	  arguments.push_back("-g");
	if( ( ioption % 100 ) / 10 )
	  arguments.push_back("-f");
      }

    }

    // 引数が整数ひとつで環境変数 SEQNEWOPT が参照できれば
    // seqnew_p から呼ばれている
    if( seqnew_p_flg ){

      fprintf(stderr, "[[ PID : %d - START    : seqnew %s ]]\n", getpid(), dinput );

      string buf = dinput;

      if( buf.find(":") == string::npos ){
	buf = trimLeftSpaces(buf);
	buf = trimRightSpaces(buf);
	arguments.push_back(buf); // データベース名
	arguments.push_back(buf); // ソースファイル名
      }
      else{
	// dbname:source1:source2:source3....
	list<string> words;
	splitToList(words,buf,":");
	list<string>::const_iterator iw = words.begin();
	while( iw != words.end() ){
	  buf = *iw++;
	  buf = trimLeftSpaces(buf);
	  buf = trimRightSpaces(buf);
	  arguments.push_back(buf);
	}
      }

      char *envdbtype = getenv("DBTYPE");
      if( envdbtype && *envdbtype ){
	buf = envdbtype;
	buf = trimLeftSpaces(buf);
	buf = trimRightSpaces(buf);
	arguments.push_back("-t");
	arguments.push_back(buf);
      }
      char *envdbpath = getenv("DBPATH");
      if( envdbpath && *envdbpath ){
	buf = envdbpath;
	buf = trimLeftSpaces(buf);
	buf = trimRightSpaces(buf);
	arguments.push_back("-d");
	arguments.push_back(buf);
      }
      char *envdbabbr = getenv("DBABBR");
      if( envdbabbr && *envdbabbr ){
	buf = envdbabbr;
	buf = trimLeftSpaces(buf);
	buf = trimRightSpaces(buf);
	arguments.push_back("-b");
	arguments.push_back(buf);
      }

    }

  }

  multimap<string,string> options;

  aitr = arguments.begin();
  while (aitr != arguments.end()) {
    if (*aitr == "-f") {
      aitr = arguments.erase(aitr);
      options.insert(make_pair("f","tit_only"));
      continue;
    }
    if (*aitr == "-g") {
      aitr = arguments.erase(aitr);
      options.insert(make_pair("g","cdb_only"));
      continue;
    }
    if (*aitr == "-cdb"||*aitr == "--cdb") {
      aitr = arguments.erase(aitr);
      options.insert(make_pair("cdb","cdb"));
      continue;
    }
    if (*aitr == "-mcdb"||*aitr == "--mcdb") {
      aitr = arguments.erase(aitr);
      options.insert(make_pair("mcdb","mcdb"));
      continue;
    }
    ++aitr;
  }


  // 引数チェックと変数セット
  list<string> option_keys;
  option_keys.push_back("h");
  option_keys.push_back("p");
  option_keys.push_back("t");
  option_keys.push_back("d");
  option_keys.push_back("b");
  option_keys.push_back("cdb");
  option_keys.push_back("mcdb");

  int r = parseargs(arguments,option_keys,options);
  if( r == 0 || r == -1 ) usage();
  if( r == 0 ) return 1;
  if( r < 0 ) return 0;

  // 引数がひとつしかない場合
  if( arguments.size () == 1 ){
    string str;
    if( cin.bad() ){
      cerr << "Standard input was failed." << endl;
      return 1;
    }

    // 標準入力が端末ならメッセージを表示
    if ( isatty( STDIN_FILENO ) )
      cerr << "Input source file(s): ";

    // EOFが来るまで読み込み
    while( getline(cin,str) ){
      if( str.empty() )
	break;

      // 標準入力が端末ならメッセージを表示
      if ( isatty( STDIN_FILENO ) )
	cerr << "Input source file(s): ";

      arguments.push_back(str);
    }

    if( arguments.size () < 2 ){
      cerr << "Source file(s) was(were) not given." << endl;
      return 1;
    }

  }

  // データベース名
  aitr = arguments.begin();
  if( aitr == arguments.end() ){
    usage();
    return 1;
  }
  string db = *aitr;
  aitr = arguments.erase(aitr);

  // T 番号以外のデータベース名はASCII小文字
  // T 番号は set_dbname(), set_abbr() で大文字変換
  Lower(db);

  bool verbose = false;
  bool debug   = false;
  bool mcdb = false;
  
#ifdef MCDB
    mcdb = true;
#endif

  pair<sci,sci> optp;

  // cdb と mcdb は同時に指定できない
  optp = options.equal_range("cdb");
  if (optp.first != optp.second) {
    mcdb = false;
    optp = options.equal_range("mcdb");
    if (optp.first != optp.second) {
      usage();
      return 1;
    }
  }
  optp = options.equal_range("mcdb");
  if (optp.first != optp.second)
    mcdb = true;
  optp = options.equal_range("v");
  if( optp.first != optp.second )
    verbose = true;
  optp = options.equal_range("D");
  if( optp.first != optp.second )
    debug = true;

  if( verbose || debug )
    while( aitr != arguments.end() )
      cout << *aitr++ << endl;

  string dbabbr;
  optp = options.equal_range("b");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbabbr = pitr->second;
      pitr++;
    }
  }
  string dbtype;
  optp = options.equal_range("t");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtype = pitr->second;
      pitr++;
    }
  }
  string dbpath;
  optp = options.equal_range("d");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbpath = pitr->second;
      pitr++;
    }
  }
  if( dbtype.empty() && ! dbpath.empty() ){
    usage();
    return 1;
  }
  if( dbtype.empty() && ! dbabbr.empty() ){
    usage();
    return 1;
  }
  if( ! dbtype.empty() && dbpath.empty() ){
    usage();
    return 1;
  }

  // dbtab ファイルの設定
  // デフォルト dbtab == BIOROOT/etc/dbtab
  string dbtab;
  char *dbtabenv = getenv("DBTAB");
  if( dbtabenv && *dbtabenv ){
    dbtab = dbtabenv;
  }
  optp = options.equal_range("p");
  if( optp.first != optp.second ){
    sci pitr = optp.first;
    while(pitr != optp.second){
      dbtab = pitr->second;
      pitr++;
    }
  }
  if( dbtab.length() == 0 ){
    dbtab = bioroot;
    dbtab += DBTAB;
  }

  // file name check
  if( basename(dbtab).length() == 0 ){
    cerr << "Invalid Input File " << dbtab << endl;
    return 1;
  }

  // dbtab のデータリスト
  list<string> inputlist;

  string proxy;
  // t オプションと d オプションが指定されている場合
  if( ! dbtype.empty() && ! dbpath.empty() ){
    // b オプションが指定されていない場合
    if( dbabbr.empty() )
      dbabbr = dbtype;
    inputlist.push_back(db + " " + dbtype + " " + dbpath + " " + dbabbr);
  }
  // dbtab の情報を取得
  else if( ! readFile(dbtab,db,inputlist,proxy) ){
    cerr << db << " : No such database in dbtab." << endl;
    return 1;
  }

  // 合成データベースとして定義されている
  // ex.) uniprot = swissprot + trembl
  if( inputlist.size() != 1 ){
    cerr << db << " : Defined as complex database in dbtab." << endl;
    return 1;
  }

  // データベース名
  string dbname;

  // データベースディレクトリ
  string dir;

  // データベースタイプ
  string type;

  // 省略名
  string abbr;

  vector<string> lbuf;
  lci iitr = inputlist.begin();

  // 半角スペース区切りの文字列をリストに変換
  splitToList(lbuf,*iitr++," ");

  vector<string>::const_iterator bitr = lbuf.begin();
  if( lbuf.size() == 4 ){
    // local database
    dbname = *bitr++;
    type = *bitr++;
    dir  = *bitr++;
    abbr = *bitr;
  }
  else if( lbuf.size() == 3 ){
    // netdbget database
    // ex.) genbank dbget.genome.ad.jp 3266
    // ex.) prosite http://www.genome.jp/dbget-bin/dbgetserv? 80
    cerr << db << " : Defined as netdbget database in dbtab." << endl;
    return 1;
  }

  int dirlen = dir.length();
  if( dir[dirlen-1] != '/' )
    dir += "/";

  // ソースファイルのファイル名のリスト
  aitr = arguments.begin();
  string givenfile;
  list<string> sourcefiles;
  while( aitr != arguments.end() ){
    givenfile = *aitr++;
    if( givenfile.find("/") != string::npos )
      givenfile = basename(givenfile);
    sourcefiles.push_back(givenfile);
  }

  aitr = sourcefiles.begin();
  if( aitr == sourcefiles.end() ){
    cerr << db << " : Flat file(s) was(were) not found." << endl;
    return 1;
  }

  Seqnew *obj;
  if( type == "genes" )
    obj = new Genes();
  else if( type == "orthology" )
    obj = new Orthology();
  else if( type == "ko" )
    obj = new Orthology();
  else if( type == "brite" )
    obj = new Brite();
  else if( type == "pathway" )
    obj = new Pathway();
  else if( type == "module" )
    obj = new Module();
  else if( type == "network" )
    obj = new Network();
  else if( type == "variant" )
    obj = new Variant();
  else if( type == "disease" )
    obj = new Disease();
  else if( type == "enzyme" )
    obj = new Enzyme();
  else if( type == "compound" )
    obj = new Compound();
  else if( type == "dgroup" )
    obj = new Dgroup();
  else if( type == "drug" )
    obj = new Drug();
  else if( type == "glycan" )
    obj = new Glycan();
  else if( type == "reaction" )
    obj = new Reaction();
  else if( type == "rclass" )
    obj = new Rclass();
  else if( type == "aaindex" )
    obj = new Aaindex();
  else if( type == "expression" )
    obj = new Expression();
  else if( type == "pdbstr" )
    obj = new Pdbstr();
  else if( type == "ncbi-cdd" )
    obj = new Ncbi_cdd();
  else if( type == "swissprot" )
    obj = new Swissprot();
  else if( type == "trembl" )
    obj = new Swissprot();
  else if( type == "refseq" )
    obj = new Refseq();
  else if( type == "pdb" )
    obj = new Pdb();
  else if( type == "carbbank" )
    obj = new Carbbank();
  else if( type == "epd" )
    obj = new Epd();
  else if( type == "pfam" )
    obj = new Pfam();
  else if( type == "pmd" )
    obj = new Pmd();
  else if( type == "prosite" )
    obj = new Prosite();
  else if( type == "prosdoc" )
    obj = new Prosdoc();
  else if( type == "sequences" )
    obj = new Sequences();
  else if( type == "genbank" )
    obj = new Genbank();
  else if( type == "embl" )
    obj = new Embl();
  else if( type == "compound_ja" )
    obj = new Compound_ja();
  else if( type == "dgroup_ja" )
    obj = new Dgroup_ja();
  else if( type == "drug_ja" )
    obj = new Drug_ja();
  else if( type == "disease_ja" )
    obj = new Disease_ja();
  else{
    cerr << "seqnew : unknown database type : " << type << endl;
    return 1;
  }

  // t オプションと d オプションが指定されている場合
  if( ! dbtype.empty() )
    dbname = db;

  obj->set_abbr(abbr);
  obj->set_dbname(dbname);
  obj->set_outdir(dir);

  // seqnew_p から呼ばれたかどうか
  obj->set_seqnew_p(false);
  if( seqnew_p_flg )
    obj->set_seqnew_p(true);

  // -f オプションは titファイルのみを生成する
  obj->set_tit_only(false);
  optp = options.equal_range("f");
  if( optp.first != optp.second )
    obj->set_tit_only(true);

  // -g オプションは cdbファイルのみを生成する
  obj->set_cdb_only(false);
  optp = options.equal_range("g");
  if( optp.first != optp.second )
    obj->set_cdb_only(true);

  // -f と -g オプションは同時に指定しない
  if( obj->_flg_tit_only & obj->_flg_cdb_only ){
    cerr << "Syntax error : cannnot specify -f and -g at the same time." << endl;
    return 1;
  }

  // -mcdb/--mcdb は mcdbファイルを生成する
  obj->set_mcdb(false);
  if (mcdb)
    obj->set_mcdb(true);

  if (!obj->open_outfiles()) {
    cerr << "Open error : output files." << endl;
    return 1;
  }

  if (!seqnew_p_flg) {
    cerr << endl;
    cerr << "           *****   IDEAS  SEQNEW   *****" << endl;
  }

  string buf;
  while (aitr != sourcefiles.end()) {
    string file = dir;
    file += *aitr;
    ifstream ifs(file.c_str());
    if (!ifs) {
      cerr << "Cannot open file: " << *aitr++ << endl;
      continue;
    }

    obj->set_flatfile(*aitr);

    // GENES についてはソースファイル名が生物種名になっている
    // その生物種ごとに省略名を取得してインデックス処理をする
    if (type == "genes") {
      string species = basename(file);
      list<string> glist;
      if (readFile(dbtab,species,glist,proxy)) {
	iitr = glist.begin();
	splitToList(lbuf,*iitr," ");
	bitr = lbuf.begin();
	if( lbuf.size() == 4 )
	  abbr = *(bitr+3);
      }
      obj->set_abbr(abbr);
      obj->set_is_genes(species,abbr);
    }

    ++aitr;

    while (getline(ifs,buf))
      obj->parser(ifs,buf);

    if (!obj->_id.empty())
      obj->writer();

    ifs.close();
  }

  obj->close_outfiles();

  if (!obj->create_cdbs()) {
    cerr << "Failed to create index." << endl;
    return 1;
  }

  if (!seqnew_p_flg) {
    cout << endl;
    cout << endl;
    cout << "           ##### SEQNEW STATISTICS #####" << endl;
    cout << "DATABASE name              : " << obj->get_dbname() << endl;
    cout << "Number of entries          : " << obj->get_counter() << endl;
    cout << "Number of rejected         : " << obj->get_rejected() << endl;
    cout << "Number of residues         : " << obj->get_total_aaseqlen() + obj->get_total_ntseqlen() << endl;
    cout << "Maximum length of sequence : " << obj->get_max_seqlen() << endl;
  }

  delete obj;

  return 0;

}

Seqnew::Seqnew():
  _counter(0),
  _rejected(0),
  _total_aaseqlen(0),
  _max_aaseqlen(0),
  _total_ntseqlen(0),
  _max_ntseqlen(0),
  _max_seqlen(0),
  _offset(0),
  _size(0),
  _id(""),
  _tit_id(""),
  _header(""),
  _title(""),
  _accession(""),
  _flatfile(""),
  _dbname(""),
  _dir(""),
  _flg_seqnew_p(false),
  _flg_tit_only(false),
  _flg_cdb_only(false),
  _is_genes(false),
  _mcdb(false){
}

Seqnew::~Seqnew(){}

void Seqnew::parser(ifstream& ifs,
		    string& buf){
}

void Seqnew::set_offset(ifstream& ifs,
			string& buf){
  _offset = ifs.tellg();
  _offset -= buf.length();
  --_offset;
}

void Seqnew::set_id(string& buf){

  // ENTRY が2回以上は不正
  if( ! _id.empty() )
    ++_rejected;
  else{
    // 半角スペースで区切られた最初の単語をキーにする
    _id += get_first_word(buf);

    // tit ファイル用の ID
    _tit_id = _abbr;
    _tit_id += ":";
    _tit_id += _id;

    ++_counter;
  }

}

void Seqnew::set_size(ifstream& ifs){

  // ENTRY が0回は不正
  if( _id.empty() )
    ++_rejected;
  else{
    _size = ifs.tellg();
    _size -= _offset;
  }

}

void Seqnew::set_seq_size(string& buf,int64_t& seq_size,int64_t& seq_offset){
  seq_size = _size - ( seq_offset - _offset );
  seq_size -= buf.length();
  --seq_size;
}

void Seqnew::set_header(string& buf){
  // 半角スペースで区切られた最初の単語をヘッダにする
  string::size_type ret = buf.find(" ");
  if( ret != buf.npos )
    buf = buf.substr(0,ret);
  _header = buf;
}

// 継続するタイトル情報を連結する
// 行末が '-'(ハイフン) でなければ半角スペースを挿入して連結
void Seqnew::extend_title(string& title,string& buf){
  if( title[title.length()-1] != '-' )
    title += " ";
  title += trimLeftSpaces(buf);
}

void Seqnew::set_seqnew_p(const bool flg){
  _flg_seqnew_p = flg;
}

void Seqnew::set_tit_only(const bool flg){
  _flg_tit_only = flg;
}

void Seqnew::set_cdb_only(const bool flg){
  _flg_cdb_only = flg;
}

void Seqnew::set_mcdb(const bool flg){
  _mcdb = flg;
}

void Seqnew::set_abbr(const string& abbr){
  _abbr = abbr;
  if( isTNumber(_abbr) || isRGNumber(_abbr) )
    Upper(_abbr);
}
string Seqnew::get_abbr(){
  return _abbr;
}

// エントリデータをインデックスファイルに追加する
void Seqnew::writer(){

  // _ofs は bget 用 base file
  // _tit は bfind 用 base file
  // _tit_sub は btit 用 base file
  // _acc は bacc 用 base file

  // -g オプションは cdbファイルのみを生成する
  if (!_flg_cdb_only)
    _tit << _tit_id << " " << _title << endl;

  // -f オプションは titファイルのみを生成する
  if (!_flg_tit_only) {
    unsigned tlen = 0;
    unsigned slen = 0;
    if (_mcdb) {
      string key = _id;
      Lower(key);
      tlen = _tit_id.length() + 1 + _title.length();
      slen = key.length() + 1;
      slen += _flatfile.length() + 1;
      slen += to_string(_offset).length() + 1;
      slen += to_string(_size).length();
      _tit_sub << "+" << key.length()
               << "," << tlen
               << ":" << key
               << "->" << _tit_id << " " << _title << endl;
      _ofs << "+" << key.length()
           << "," << slen
           << ":" << key
           << "->" << _id
           << " " << _flatfile << " " << _offset << " " << _size << endl;
    }
    else {
      _tit_sub << _id << " " << _tit_id << " " << _title << endl;
      _ofs << _id << " " << _flatfile << " " << _offset << " " << _size << endl;
    }

    if (!_accession.empty()) {
      //_acc << _id << " " << _accession << endl;

      // 複数のアクセッションのうち、
      // プライマリアクセッションのみをキーとして利用する
      list<string> lacc;
      splitToList(lacc,_accession," ");
      lci itr = lacc.begin();
      string _primary_acc = *itr;

      // ID と AC が同じだったら出力しない
      if (_primary_acc != _id) {
        if (_mcdb) {
          string key = _primary_acc;
          Lower(key);
          slen = key.length() + 1;
          slen += _flatfile.length() + 1;
          slen += to_string(_offset).length() + 1;
          slen += to_string(_size).length();
          _tit_sub << "+" << key.length()
                   << "," << tlen
                   << ":" << key
                   << "->" << _tit_id << " " << _title << endl;
          _ofs << "+" << key.length()
               << "," << slen
               << ":" << key
               << "->" << _primary_acc
               << " " << _flatfile << " " << _offset << " " << _size << endl;
        }
        else {
          _tit_sub << _primary_acc << " "  << _tit_id << " " << _title << endl;
          //_acc << _primary_acc << " " << _accession << endl;
          _ofs << _primary_acc << " " << _flatfile << " " << _offset << " " << _size << endl;
        }
      }
    }
  }

  // エントリ数カウンタ
  if (!_flg_seqnew_p) {
    if ((_counter % 1000) == 0) cerr << ".";
    if ((_counter % 50000) == 0) cerr << endl;
  }

}

// 配列データや KCF データを bget用インデックスファイルに追加する
void Seqnew::writer(const string& _data_id,
		    const int64_t& _data_offset,
		    const int64_t& _data_size){

  // -f オプションは titファイルのみを生成する
  if (_flg_tit_only)
    return;

  unsigned slen = 0;
  if (_mcdb) {
    string key = _data_id;
    Lower(key);
    slen = key.length() + 1;
    slen += _flatfile.length() + 1;
    slen += to_string(_data_offset).length() + 1;
    slen += to_string(_data_size).length();
    _ofs << "+" << key.length()
         << "," << slen
         << ":" << key
         << "->" << _data_id
         << " " << _flatfile << " " << _data_offset
         << " " << _data_size << endl;
  }
  else {
    _ofs << _data_id << " " << _flatfile << " "
         << _data_offset << " " << _data_size << endl;
  }

  if (!_accession.empty()) {
    // 複数のアクセッションのうち、
    // プライマリアクセッションのみをキーとして利用する
    list<string> lacc;
    splitToList(lacc,_accession," ");
    lci itr = lacc.begin();
    // _data_id == _aaseq_xxxx, _ntseq_xxxx or _kcf_xxxx 
    string::size_type ret = _data_id.find("_",1);
    if (ret == _data_id.npos)
      return;
    string _primary_acc = _data_id.substr(0,ret+1);
    _primary_acc += *itr;

    // ID と AC が同じだったら出力しない
    if (_primary_acc != _data_id) {
      if (_mcdb) {
        string key = _primary_acc;
        Lower(key);
        slen = key.length() + 1;
        slen += _flatfile.length() + 1;
        slen += to_string(_data_offset).length() + 1;
        slen += to_string(_data_size).length();
        _ofs << "+" << key.length()
             << "," << slen
             << ":" << key
             << "->" << _primary_acc
             << " " << _flatfile << " " << _data_offset
             << " " << _data_size << endl;
      }
      else {
        _ofs << _primary_acc << " " << _flatfile << " "
             << _data_offset << " " << _data_size << endl;
      }
    }
  }
}

void Seqnew::clear(){
  _offset = 0;
  _size = 0;
  _id = "";
  _tit_id = "";
  _header = "";
  _title = "";
  _accession = "";
}

void Seqnew::set_is_genes(string& file,string& abbr){};

unsigned long Seqnew::get_counter(){ return _counter; }
unsigned long Seqnew::get_rejected(){ return _rejected; }
int64_t Seqnew::get_total_aaseqlen(){ return _total_aaseqlen; }
int64_t Seqnew::get_max_aaseqlen(){ return _max_aaseqlen; }
int64_t Seqnew::get_total_ntseqlen(){ return _total_ntseqlen; }
int64_t Seqnew::get_max_ntseqlen(){ return _max_ntseqlen; }
int64_t Seqnew::get_max_seqlen(){ return _max_seqlen; }

void Seqnew::set_flatfile(const string& flatfile){
  _flatfile = flatfile;
}
void Seqnew::set_dbname(const string& dbname){
  _dbname = dbname;
  if( isTNumber(_dbname) || isRGNumber(_dbname) )
    Upper(_dbname);
}
string Seqnew::get_dbname(){
  return _dbname;
}

string Seqnew::get_first_word(const string& str){
  string buf = trimLeftSpaces(str);
  string::size_type ret = buf.find(" ");
  if( ret != buf.npos )
    buf = buf.substr(0,ret);
  return buf;
}

void Seqnew::set_outdir(const string& dir){
  _dir = dir;
}

bool Seqnew::open_outfiles(){

  string outfile = _dir;
  outfile += _dbname;

  // .bin は bget 用 base file
  // .tit は bfind 用 base file
  // .tit_sub は btit 用 base file
  // .acc は bacc 用 base file

  string posfile = outfile;
  posfile += ".bin";
  string titfile = outfile;
  titfile += ".tit";
  string titsubfile = outfile;
  titsubfile += ".tit_sub";
  string accfile = outfile;
  accfile += ".acc";

  // -g オプションは cdbファイルのみを生成する
  if( ! _flg_cdb_only ){
    _tit.open(titfile.c_str());

    if( !_tit ){
      cerr << "Cannot open file: " << titfile << endl;
      return false;
    }
  }

  // -f オプションは titファイルのみを生成する
  if( _flg_tit_only )
    return true;

  _ofs.open(posfile.c_str());
  _tit_sub.open(titsubfile.c_str());
  //_acc.open(accfile.c_str());

  if( !_ofs ){
    cerr << "Cannot open file: " << posfile << endl;
    return false;
  }
  if( !_tit_sub ){
    cerr << "Cannot open file: " << titsubfile << endl;
    return false;
  }
  //if( !_acc ){
  //  cerr << "Cannot open file: " << accfile << endl;
  //  return false;
  //}

  return true;

}
void Seqnew::close_outfiles(){

  // -g オプションは cdbファイルのみを生成する
  if (!_flg_cdb_only)
    _tit.close();

  // -f オプションは titファイルのみを生成する
  if (_flg_tit_only)
    return;

  if (_mcdb) {
    _ofs << "+0,0:->" << endl;
    _ofs << endl;
    _tit_sub << "+0,0:->" << endl;
    _tit_sub << endl;
  }

  _ofs.close();
  _tit_sub.close();
  //_acc.close();
}
bool Seqnew::create_cdbs(){

  // -f オプションは titファイルのみを生成する
  if (_flg_tit_only)
    return true;

  string outfile = _dir;
  outfile += _dbname;

  // 作成モード 0
  // 半角スペースで分割した最初の単語をキーにする
  // データは行全体

  // 作成モード 1
  // 半角スペースで分割した最初の単語をキーにする
  // データはキーの後ろに続く文字列

  // エントリデータ
  string basefile = outfile;
  basefile += ".bin";

  // title データ
  string tsubfile = outfile;
  tsubfile += ".tit_sub";

  string cdbfile = outfile;
  if (_mcdb)
    cdbfile += ".mcdb";
  else
    cdbfile += ".cdb";

  // 作成モード 0
  bool flg = false;
  if (_mcdb)
    flg = create_mcdb(basefile,cdbfile);
  else
    flg = create_cdb(basefile,cdbfile,0);
  if (!flg) {
    unlink(basefile.c_str());
    unlink(tsubfile.c_str());
    return false;
  }

  unlink(basefile.c_str());

  cdbfile = outfile;
  if (_mcdb)
    cdbfile += ".tit.mcdb";
  else
    cdbfile += ".tit.cdb";

  // 作成モード 1
  if (_mcdb)
    flg = create_mcdb(tsubfile,cdbfile);
  else
    flg =  create_cdb(tsubfile,cdbfile,1);
  if (!flg) {
    unlink(tsubfile.c_str());
    return false;
  }

  unlink(tsubfile.c_str());

  // accession データ
  basefile = outfile;
  basefile += ".acc";

  cdbfile = outfile;
  cdbfile += ".acc.cdb";

  // 作成モード 1
  //if( ! create_cdb(basefile,cdbfile,1) ){
  //  unlink(basefile.c_str());
  //  return false;
  //}

  //unlink(basefile.c_str());

  return true;

}

bool Seqnew::create_cdb(const string& basefile,
			const string& cdbfile,
			int mode){
  string buf;
  list<string> lbuf;
  ifstream ifs(basefile.c_str());
  if (!ifs)
    cerr << basefile << " : open error." << endl;

  int fd;
  struct cdb_make cdbm;
  unsigned keylen, datalen;

  // CDB を初期化
  fd = open(cdbfile.c_str(), O_RDWR|O_CREAT|O_TRUNC, 0644);
  if( fd < 0 ){
    cerr << cdbfile << " : open failed." << endl;
    return false;
  }

  if( cdb_make_start(&cdbm, fd) < 0 ){
    cerr << "Failed to initialize structure to create a database." << endl;
    return false;
  }

  while( getline(ifs,buf) ) {

    // 半角スペースで区切られた最初の単語をキーにする
    string ikey;
    string::size_type ret = buf.find(" ");
    if( ret == buf.npos )
      continue;
    ikey = buf.substr(0,ret);

    // データは行全体
    // mode == 0 デフォルト .cdb 用の処理

    // データはキーの後ろに続く文字列
    // .tit.cdb、.acc.cdb 用の処理
    if( mode == 1 )
      buf = buf.substr(ret+1);

    // インデックス用キーは小文字
    Lower(ikey);
    keylen = ikey.length();
    datalen = buf.length();

    // cdb_make_exists() may significantly slow down the whole process,
    // and currently it flushes internal buffer to disk on every call
    // with key those hash value already exists in db.
    //if (cdb_make_exists(&cdbm, ikey.c_str(), keylen) == 0)
    if( cdb_make_add(&cdbm, ikey.c_str(), keylen, buf.c_str(), datalen) < 0 ){
      cerr << "Failed to add record to database." << endl;
      return false;
    }

    // or use cdb_make_put() with appropriate flags

  }

  // finalize and close the database
  if( cdb_make_finish(&cdbm) < 0 ){
    cerr << "Failed to finalize or close the database." << endl;
    return false;
  }
  close(fd);

  fd = open(cdbfile.c_str(), O_RDONLY);
  if( fd < 0 ){
    //cerr << cdbfile << " : Open error." << endl;
    return false;
  }

  // インデックスファイルのサイズチェック
  struct stat st;
  fstat(fd,&st);
  if( st.st_size >= UINT_MAX ){
    cerr << "Error: " << cdbfile << " exceeds size limit." << endl;
    return false;
  }
  close(fd);

  return true;
}

bool Seqnew::create_mcdb(const string& basefile,
                         const string& cdbfile){
  const char *fname = cdbfile.c_str();
  const char *input = basefile.c_str();
  if (mcdb_makefmt_fileintofile(input, fname, malloc, free) < 0) {
    cerr << "mcdb_makefmt_fileintofile() : Error." << endl;
    return false;
  }
  return true;
}
